# update_maintainer.py - updates the Maintainer field of an Ubuntu package
#
# Copyright (C) 2010, Benjamin Drung <bdrung@ubuntu.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""This module is for updating the Maintainer field of an Ubuntu package."""

import logging
import os
import re

import debian.changelog

Logger = logging.getLogger(__name__)

# Prior May 2009 these Maintainers were used:
_PREVIOUS_UBUNTU_MAINTAINER = (
    "ubuntu core developers <ubuntu-devel@lists.ubuntu.com>",
    "ubuntu core developers <ubuntu-devel-discuss@lists.ubuntu.com>",
    "ubuntu motu developers <ubuntu-motu@lists.ubuntu.com>",
)
_UBUNTU_MAINTAINER = "Ubuntu Developers <ubuntu-devel-discuss@lists.ubuntu.com>"


class MaintainerUpdateException(Exception):
    pass


class Control:
    """Represents a debian/control file"""

    def __init__(self, filename):
        assert os.path.isfile(filename), f"{filename} does not exist."
        self._filename = filename
        self._content = open(filename, encoding="utf-8").read()

    def get_maintainer(self):
        """Returns the value of the Maintainer field."""
        maintainer = re.search("^Maintainer: ?(.*)$", self._content, re.MULTILINE)
        if maintainer:
            maintainer = maintainer.group(1)
        return maintainer

    def get_original_maintainer(self):
        """Returns the value of the XSBC-Original-Maintainer field."""
        orig_maintainer = re.search(
            "^(?:[XSBC]*-)?Original-Maintainer: ?(.*)$", self._content, re.MULTILINE
        )
        if orig_maintainer:
            orig_maintainer = orig_maintainer.group(1)
        return orig_maintainer

    def save(self, filename=None):
        """Saves the control file."""
        if filename:
            self._filename = filename
        control_file = open(self._filename, "w", encoding="utf-8")
        control_file.write(self._content)
        control_file.close()

    def set_maintainer(self, maintainer):
        """Sets the value of the Maintainer field."""
        pattern = re.compile("^Maintainer: ?.*$", re.MULTILINE)
        self._content = pattern.sub(f"Maintainer: {maintainer}", self._content)

    def set_original_maintainer(self, original_maintainer):
        """Sets the value of the XSBC-Original-Maintainer field."""
        original_maintainer = f"XSBC-Original-Maintainer: {original_maintainer}"
        if self.get_original_maintainer():
            pattern = re.compile("^(?:[XSBC]*-)?Original-Maintainer:.*$", re.MULTILINE)
            self._content = pattern.sub(original_maintainer, self._content)
        else:
            pattern = re.compile("^(Maintainer:.*)$", re.MULTILINE)
            self._content = pattern.sub(f"\\1\\n{original_maintainer}", self._content)

    def remove_original_maintainer(self):
        """Strip out out the XSBC-Original-Maintainer line"""
        pattern = re.compile(
            "^(?:[XSBC]*-)?Original-Maintainer:.*?$.*?^", re.MULTILINE | re.DOTALL
        )
        self._content = pattern.sub("", self._content)


def _get_distribution(changelog_file):
    """get distribution of latest changelog entry"""
    changelog = debian.changelog.Changelog(
        open(changelog_file, encoding="utf-8"), strict=False, max_blocks=1
    )
    distribution = changelog.distributions.split()[0]
    # Strip things like "-proposed-updates" or "-security" from distribution
    return distribution.split("-", 1)[0]


def _find_files(debian_directory, verbose):
    """Find possible control files.
    Returns (changelog, control files list)
    Raises an exception if none can be found.
    """
    possible_contol_files = [os.path.join(debian_directory, f) for f in ["control.in", "control"]]

    changelog_file = os.path.join(debian_directory, "changelog")
    control_files = [f for f in possible_contol_files if os.path.isfile(f)]

    # Make sure that a changelog and control file is available
    if len(control_files) == 0:
        raise MaintainerUpdateException(f"No control file found in {debian_directory}.")
    if not os.path.isfile(changelog_file):
        raise MaintainerUpdateException(f"No changelog file found in {debian_directory}.")

    # If the rules file accounts for XSBC-Original-Maintainer, we should not
    # touch it in this package (e.g. the python package).
    rules_file = os.path.join(debian_directory, "rules")
    if (
        os.path.isfile(rules_file)
        and "XSBC-Original-" in open(rules_file, encoding="utf-8").read()
    ):
        if verbose:
            print("XSBC-Original is managed by 'rules' file. Doing nothing.")
        control_files = []

    return (changelog_file, control_files)


def update_maintainer(debian_directory, verbose=False):
    """updates the Maintainer field of an Ubuntu package

    * No modifications are made if the Maintainer field contains an ubuntu.com
      email address. Otherwise, the Maintainer field will be set to
      Ubuntu Developers <ubuntu-devel-discuss@lists.ubuntu.com>
    * The old value will be saved in a field named XSBC-Original-Maintainer
      if the Maintainer field is modified.

    Policy: https://wiki.ubuntu.com/DebianMaintainerField
    """
    try:
        changelog_file, control_files = _find_files(debian_directory, verbose)
    except MaintainerUpdateException as e:
        Logger.error(str(e))
        raise

    distribution = _get_distribution(changelog_file)
    for control_file in control_files:
        control = Control(control_file)
        original_maintainer = control.get_maintainer()

        if original_maintainer is None:
            Logger.error("No Maintainer field found in %s.", control_file)
            raise MaintainerUpdateException("No Maintainer field found")

        if original_maintainer.strip().lower() in _PREVIOUS_UBUNTU_MAINTAINER:
            if verbose:
                print(f"The old maintainer was: {original_maintainer}")
                print(f"Resetting as: {_UBUNTU_MAINTAINER}")
            control.set_maintainer(_UBUNTU_MAINTAINER)
            control.save()
            continue

        if original_maintainer.strip().endswith("ubuntu.com>"):
            if verbose:
                print("The Maintainer email is set to an ubuntu.com address. Doing nothing.")
            continue

        if distribution in ("stable", "testing", "unstable", "experimental"):
            if verbose:
                print("The package targets Debian. Doing nothing.")
            return

        if control.get_original_maintainer() is not None:
            Logger.warning(
                "Overwriting original maintainer: %s", control.get_original_maintainer()
            )

        if verbose:
            print(f"The original maintainer is: {original_maintainer}")
            print(f"Resetting as: {_UBUNTU_MAINTAINER}")
        control.set_original_maintainer(original_maintainer)
        control.set_maintainer(_UBUNTU_MAINTAINER)
        control.save()

    return


def restore_maintainer(debian_directory, verbose=False):
    """Restore the original maintainer"""
    try:
        control_files = _find_files(debian_directory, verbose)[1]
    except MaintainerUpdateException as e:
        Logger.error(str(e))
        raise

    for control_file in control_files:
        control = Control(control_file)
        orig_maintainer = control.get_original_maintainer()
        if not orig_maintainer:
            continue
        if verbose:
            print(f"Restoring original maintainer: {orig_maintainer}")
        control.set_maintainer(orig_maintainer)
        control.remove_original_maintainer()
        control.save()



#
# source_package.py - Internal helper class for sponsor-patch
#
# Copyright (C) 2011, Benjamin Drung <bdrung@ubuntu.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import logging
import os
import pathlib
import re
import subprocess
import sys

import debian.changelog
import debian.deb822

from ubuntutools.question import Question, YesNoQuestion
from ubuntutools.sponsor_patch.question import (
    ask_for_ignoring_or_fixing,
    ask_for_manual_fixing,
    user_abort,
)

Logger = logging.getLogger(__name__)


def _get_series(launchpad):
    """Returns a tuple with the development and list of supported series."""
    ubuntu = launchpad.distributions["ubuntu"]
    devel_series = ubuntu.current_series.name
    supported_series = [
        series.name for series in ubuntu.series if series.active and series.name != devel_series
    ]
    return (devel_series, supported_series)


def strip_epoch(version):
    """Removes the epoch from a Debian version string.

    strip_epoch(1:1.52-1) will return "1.52-1" and strip_epoch(1.1.3-1) will
    return "1.1.3-1".
    """

    parts = version.full_version.split(":")
    if len(parts) > 1:
        del parts[0]
    version_without_epoch = ":".join(parts)
    return version_without_epoch


class SourcePackage:
    """This class represents a source package."""

    def __init__(self, package, builder, workdir, branch):
        self._package = package
        self._builder = builder
        self._workdir = workdir
        self._branch = branch
        self._changelog = None
        self._version = None
        self._build_log = None

    def ack_sync(self, upload, task, launchpad):
        """Acknowledge a sync request and subscribe ubuntu-archive."""

        if upload == "ubuntu":
            self._print_logs()
            question = Question(["yes", "edit", "no"])
            answer = question.ask("Do you want to acknowledge the sync request", "no")
            if answer == "edit":
                return False
            if answer == "no":
                user_abort()

            bug = task.bug
            task.status = "Confirmed"
            if task.importance == "Undecided":
                task.importance = "Wishlist"
            task.lp_save()
            Logger.debug("Set bug #%i status to Confirmed.", bug.id)

            msg = "Sync request ACK'd."
            if self._build_log:
                msg = (
                    f"{self._package} {self._version} builds"
                    f" on {self._builder.get_architecture()}. {msg}"
                )
            bug.newMessage(content=msg, subject="sponsor-patch")
            Logger.debug("Acknowledged sync request bug #%i.", bug.id)

            bug.subscribe(person=launchpad.people["ubuntu-archive"])
            Logger.debug("Subscribed ubuntu-archive to bug #%i.", bug.id)

            bug.subscribe(person=launchpad.me)
            Logger.debug("Subscribed me to bug #%i.", bug.id)

            sponsorsteam = launchpad.people["ubuntu-sponsors"]
            for sub in bug.subscriptions:
                if sub.person == sponsorsteam and sub.canBeUnsubscribedByUser():
                    bug.unsubscribe(person=launchpad.people["ubuntu-sponsors"])
                    Logger.debug("Unsubscribed ubuntu-sponsors from bug #%i.", bug.id)
                elif sub.person == sponsorsteam:
                    Logger.debug("Couldn't unsubscribe ubuntu-sponsors from bug #%i.", bug.id)

            Logger.info("Successfully acknowledged sync request bug #%i.", bug.id)
        else:
            Logger.error(
                "Sync requests can only be acknowledged when the upload target is Ubuntu."
            )
            sys.exit(1)
        return True

    def ask_and_upload(self, upload):
        """Ask the user before uploading the source package.

        Returns true if the source package is uploaded successfully. Returns
        false if the user wants to change something.
        """

        # Upload package
        if upload:
            self._print_logs()
            if upload == "ubuntu":
                target = "the official Ubuntu archive"
            else:
                target = upload
            question = Question(["yes", "edit", "no"])
            answer = question.ask(f"Do you want to upload the package to {target}", "no")
            if answer == "edit":
                return False
            if answer == "no":
                user_abort()
            cmd = ["dput", "--force", upload, self._changes_file]
            Logger.debug(" ".join(cmd))
            if subprocess.call(cmd) != 0:
                Logger.error(
                    "Upload of %s to %s failed.", os.path.basename(self._changes_file), upload
                )
                sys.exit(1)

            # Push the branch if the package is uploaded to the Ubuntu archive.
            if upload == "ubuntu" and self._branch:
                cmd = ["debcommit"]
                Logger.debug(" ".join(cmd))
                if subprocess.call(cmd) != 0:
                    Logger.error("Bzr commit failed.")
                    sys.exit(1)
                cmd = ["bzr", "mark-uploaded"]
                Logger.debug(" ".join(cmd))
                if subprocess.call(cmd) != 0:
                    Logger.error("Bzr tagging failed.")
                    sys.exit(1)
                cmd = ["bzr", "push", ":parent"]
                Logger.debug(" ".join(cmd))
                if subprocess.call(cmd) != 0:
                    Logger.error("Bzr push failed.")
                    sys.exit(1)
        return True

    def build(self, update, dist=None):
        """Tries to build the package.

        Returns true if the package was built successfully. Returns false
        if the user wants to change something.
        """

        if dist is None:
            dist = re.sub("-.*$", "", self._changelog.distributions)
        build_name = (
            f"{self._package}_{strip_epoch(self._version)}"
            f"_{self._builder.get_architecture()}.build"
        )
        self._build_log = os.path.join(self._buildresult, build_name)

        successful_built = False
        while not successful_built:
            if update:
                ret = self._builder.update(dist)
                if ret != 0:
                    ask_for_manual_fixing()
                    break
                # We want to update the build environment only once, but not
                # after every manual fix.
                update = False

            # build package
            result = self._builder.build(self._dsc_file, dist, self._buildresult)
            if result != 0:
                question = Question(["yes", "update", "retry", "no"])
                answer = question.ask("Do you want to resolve this issue manually", "yes")
                if answer == "yes":
                    break
                if answer == "update":
                    update = True
                    continue
                if answer == "retry":
                    continue
                user_abort()
            successful_built = True
        if not successful_built:
            # We want to do a manual fix if the build failed.
            return False
        return True

    @property
    def _buildresult(self):
        """Returns the directory for the build result."""
        return os.path.join(self._workdir, "buildresult")

    def build_source(self, keyid, upload, previous_version):
        """Tries to build the source package.

        Returns true if the source package was built successfully. Returns false
        if the user wants to change something.
        """

        if self._branch:
            cmd = ["bzr", "builddeb", "--builder=debuild", "-S", "--", "--no-lintian", "-nc"]
        else:
            cmd = ["debuild", "--no-lintian", "-nc", "-S"]
        cmd.append("-v" + previous_version.full_version)
        if (
            previous_version.upstream_version == self._changelog.upstream_version
            and upload == "ubuntu"
        ):
            # FIXME: Add proper check that catches cases like changed
            # compression (.tar.gz -> tar.bz2) and multiple orig source tarballs
            cmd.append("-sd")
        else:
            cmd.append("-sa")
        if keyid is not None:
            cmd += ["-k" + keyid]
        env = os.environ
        if upload == "ubuntu":
            env["DEB_VENDOR"] = "Ubuntu"
        Logger.debug(" ".join(cmd))
        if subprocess.call(cmd, env=env) != 0:
            Logger.error("Failed to build source tarball.")
            # TODO: Add a "retry" option
            ask_for_manual_fixing()
            return False
        return True

    @property
    def _changes_file(self):
        """Returns the file name of the .changes file."""
        return os.path.join(
            self._workdir, f"{self._package}_{strip_epoch(self._version)}_source.changes"
        )

    def check_target(self, upload, launchpad):
        """Make sure that the target is correct.

        Returns true if the target is correct. Returns false if the user
        wants to change something.
        """

        devel_series, supported_series = _get_series(launchpad)

        if upload == "ubuntu":
            allowed = (
                supported_series + [s + "-proposed" for s in supported_series] + [devel_series]
            )
            if self._changelog.distributions not in allowed:
                Logger.error(
                    "%s is not an allowed series. It needs to be one of %s.",
                    self._changelog.distributions,
                    ", ".join(allowed),
                )
                return ask_for_ignoring_or_fixing()
        elif upload and upload.startswith("ppa/"):
            allowed = supported_series + [devel_series]
            if self._changelog.distributions not in allowed:
                Logger.error(
                    "%s is not an allowed series. It needs to be one of %s.",
                    self._changelog.distributions,
                    ", ".join(allowed),
                )
                return ask_for_ignoring_or_fixing()
        return True

    def check_version(self, previous_version):
        """Check if the version of the package is greater than the given one.

        Return true if the version of the package is newer. Returns false
        if the user wants to change something.
        """

        if self._version <= previous_version:
            Logger.error(
                "The version %s is not greater than the already available %s.",
                self._version,
                previous_version,
            )
            return ask_for_ignoring_or_fixing()
        return True

    def check_sync_request_version(self, bug_number, task):
        """Check if the downloaded version of the package is mentioned in the
        bug title."""

        if not task.title_contains(self._version):
            print(f"Bug #{bug_number} title: {task.get_bug_title()}")
            msg = f"Is {self._package} {self._version} the version that should be synced"
            answer = YesNoQuestion().ask(msg, "no")
            if answer == "no":
                user_abort()

    @property
    def _debdiff_filename(self):
        """Returns the file name of the .debdiff file."""
        debdiff_name = f"{self._package}_{strip_epoch(self._version)}.debdiff"
        return os.path.join(self._workdir, debdiff_name)

    @property
    def _dsc_file(self):
        """Returns the file name of the .dsc file."""
        return os.path.join(self._workdir, f"{self._package}_{strip_epoch(self._version)}.dsc")

    def generate_debdiff(self, dsc_file):
        """Generates a debdiff between the given .dsc file and this source
        package."""

        assert os.path.isfile(dsc_file), f"{dsc_file} does not exist."
        assert os.path.isfile(self._dsc_file), f"{self._dsc_file} does not exist."
        cmd = ["debdiff", dsc_file, self._dsc_file]
        if not Logger.isEnabledFor(logging.DEBUG):
            cmd.insert(1, "-q")
        Logger.debug("%s > %s", " ".join(cmd), self._debdiff_filename)
        with open(self._debdiff_filename, "w", encoding="utf-8") as debdiff_file:
            debdiff = subprocess.run(cmd, check=False, stdout=debdiff_file)
            assert debdiff.returncode in (0, 1)

    def is_fixed(self, lp_bug):
        """Make sure that the given Launchpad bug is closed.

        Returns true if the bug is closed. Returns false if the user wants to
        change something.
        """

        assert os.path.isfile(self._changes_file), f"{self._changes_file} does not exist."
        changes = debian.deb822.Changes(open(self._changes_file, encoding="utf-8"))
        fixed_bugs = []
        if "Launchpad-Bugs-Fixed" in changes:
            fixed_bugs = changes["Launchpad-Bugs-Fixed"].split(" ")
        fixed_bugs = [int(bug) for bug in fixed_bugs]

        # Check if the given bug is marked as duplicate. In this case get the
        # master bug.
        while lp_bug.duplicate_of:
            lp_bug = lp_bug.duplicate_of

        if lp_bug.id not in fixed_bugs:
            Logger.error("Launchpad bug #%i is not closed by new version.", lp_bug.id)
            return ask_for_ignoring_or_fixing()
        return True

    def _print_logs(self):
        """Print things that should be checked before uploading a package."""

        lintian_filename = self._run_lintian()
        print(f"\nPlease check {self._package} {self._version} carefully:")
        if os.path.isfile(self._debdiff_filename):
            print("file://" + self._debdiff_filename)
        print("file://" + lintian_filename)
        if self._build_log:
            print("file://" + self._build_log)

    def reload_changelog(self):
        """Reloads debian/changelog and updates the version.

        Returns true if the changelog was reloaded successfully. Returns false
        if the user wants to correct a broken changelog.
        """

        # Check the changelog
        self._changelog = debian.changelog.Changelog()
        try:
            self._changelog.parse_changelog(
                open("debian/changelog", encoding="utf-8"), max_blocks=1, strict=True
            )
        except debian.changelog.ChangelogParseError as error:
            Logger.error("The changelog entry doesn't validate: %s", str(error))
            ask_for_manual_fixing()
            return False

        # Get new version of package
        try:
            self._version = self._changelog.get_version()
        except IndexError:
            Logger.error(
                "Debian package version could not be determined. "
                "debian/changelog is probably malformed."
            )
            ask_for_manual_fixing()
            return False

        return True

    def _run_lintian(self) -> str:
        """Runs lintian on either the source or binary changes file.

        Returns the filename of the created lintian output file.
        """

        # Determine whether to use the source or binary build for lintian
        package_and_version = f"{self._package}_{strip_epoch(self._version)}"
        if self._build_log:
            build_changes = f"{package_and_version}_{self._builder.get_architecture()}.changes"
            changes_for_lintian = os.path.join(self._buildresult, build_changes)
        else:
            changes_for_lintian = self._changes_file

        # Check lintian
        assert os.path.isfile(changes_for_lintian), f"{changes_for_lintian} does not exist."
        cmd = ["lintian", "-IE", "--pedantic", "-q", "--profile", "ubuntu", changes_for_lintian]
        lintian_file = pathlib.Path(self._workdir) / f"{package_and_version}.lintian"
        Logger.debug("%s > %s", " ".join(cmd), lintian_file)
        with lintian_file.open("wb") as outfile:
            subprocess.run(cmd, stdout=outfile, check=True)

        return str(lintian_file)

    def sync(self, upload, series, bug_number, requester):
        """Does a sync of the source package."""

        if upload == "ubuntu":
            cmd = [
                "syncpackage",
                self._package,
                "-b",
                str(bug_number),
                "-f",
                "-s",
                requester,
                "-V",
                str(self._version),
                "-d",
                series,
            ]
            Logger.debug(" ".join(cmd))
            if subprocess.call(cmd) != 0:
                Logger.error("Syncing of %s %s failed.", self._package, str(self._version))
                sys.exit(1)
        else:
            # FIXME: Support this use case!
            Logger.error("Uploading a synced package other than to Ubuntu is not supported yet!")
            sys.exit(1)
        return True



#
# question.py - Helper class for asking questions
#
# Copyright (C) 2010, Benjamin Drung <bdrung@ubuntu.com>,
#               2011, Stefano Rivera <stefanor@ubuntu.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import os
import re
import subprocess
import sys
import tempfile


class Question:
    def __init__(self, options, show_help=True):
        assert len(options) >= 2
        self.options = [s.lower() for s in options]
        self.show_help = show_help

    def get_options(self):
        if len(self.options) == 2:
            options = f"{self.options[0]} or {self.options[1]}"
        else:
            options = f"{', '.join(self.options[:-1])}, or {self.options[-1]}"
        return options

    def ask(self, question, default=None):
        if default is None:
            default = self.options[0]
        assert default in self.options

        separator = " ["
        for option in self.options:
            if option == default:
                question += separator + option[0].upper()
            else:
                question += separator + option[0]
            separator = "|"
        if self.show_help:
            question += "|?"
        question += "]? "

        selected = None
        while selected not in self.options:
            try:
                selected = input(question).strip().lower()
            except (EOFError, KeyboardInterrupt):
                print("\nAborting as requested.")
                sys.exit(1)
            if selected == "":
                selected = default
            else:
                for option in self.options:
                    # Example: User typed "y" instead of "yes".
                    if selected == option[0]:
                        selected = option
            if selected not in self.options:
                print(f"Please answer the question with {self.get_options()}.")
        return selected


class YesNoQuestion(Question):
    def __init__(self):
        Question.__init__(self, ["yes", "no"], False)


def input_number(question, min_number, max_number, default=None):
    if default:
        question += f" [{default}]? "
    else:
        question += "? "
    selected = None
    while not selected or selected < min_number or selected > max_number:
        try:
            selected = input(question).strip()
        except (EOFError, KeyboardInterrupt):
            print("\nAborting as requested.")
            sys.exit(1)
        if default and selected == "":
            selected = default
        else:
            try:
                selected = int(selected)
                if selected < min_number or selected > max_number:
                    print(f"Please input a number between {min_number} and {max_number}.")
            except ValueError:
                print("Please input a number.")
    assert isinstance(selected, int)
    return selected


def confirmation_prompt(message=None, action=None):
    """Display message, or a stock message including action, and wait for the
    user to press Enter
    """
    if message is None:
        if action is None:
            action = "continue"
        message = f"Press [Enter] to {action}. Press [Ctrl-C] to abort now."
    try:
        input(message)
    except (EOFError, KeyboardInterrupt):
        print("\nAborting as requested.")
        sys.exit(1)


class EditFile:
    def __init__(self, filename, description, placeholders=None):
        self.filename = filename
        self.description = description
        if placeholders is None:
            placeholders = (re.compile(r"^>>>.*<<<$", re.UNICODE),)
        self.placeholders = placeholders

    def edit(self, optional=False):
        if optional:
            print(f"\n\nCurrently the {self.description} looks like:")
            with open(self.filename, "r", encoding="utf-8") as f:
                print(f.read())
            if YesNoQuestion().ask("Edit", "no") == "no":
                return

        done = False
        while not done:
            old_mtime = os.stat(self.filename).st_mtime
            subprocess.check_call(["sensible-editor", self.filename])
            modified = old_mtime != os.stat(self.filename).st_mtime
            placeholders_present = False
            if self.placeholders:
                with open(self.filename, "r", encoding="utf-8") as f:
                    for line in f:
                        for placeholder in self.placeholders:
                            if placeholder.search(line.strip()):
                                placeholders_present = True

            if placeholders_present:
                print(
                    f"Placeholders still present in the {self.description}. "
                    f"Please replace them with useful information."
                )
                confirmation_prompt(action="edit again")
            elif not modified:
                print(f"The {self.description} was not modified")
                if YesNoQuestion().ask("Edit again", "yes") == "no":
                    done = True
            elif self.check_edit():
                done = True

    def check_edit(self):  # pylint: disable=no-self-use
        """Override this to implement extra checks on the edited report.
        Should return False if another round of editing is needed,
        and should prompt the user to confirm that, if necessary.
        """
        return True


class EditBugReport(EditFile):
    split_re = re.compile(r"^Summary.*?:\s+(.*?)\s+Description:\s+(.*)$", re.DOTALL | re.UNICODE)

    def __init__(self, subject, body, placeholders=None):
        prefix = f"{os.path.basename(sys.argv[0])}_"
        tmpfile = tempfile.NamedTemporaryFile(prefix=prefix, suffix=".txt", delete=False)
        tmpfile.write((f"Summary (one line):\n{subject}\n\nDescription:\n{body}").encode("utf-8"))
        tmpfile.close()
        super().__init__(tmpfile.name, "bug report", placeholders)

    def check_edit(self):
        with open(self.filename, "r", encoding="utf-8") as f:
            report = f.read()

        if self.split_re.match(report) is None:
            print(
                f"The {self.description} doesn't start with 'Summary:' and 'Description:' blocks"
            )
            confirmation_prompt("edit again")
            return False
        return True

    def get_report(self):
        with open(self.filename, "r", encoding="utf-8") as f:
            report = f.read()

        match = self.split_re.match(report)
        title = match.group(1).replace("\n", " ")
        report = (title, match.group(2))
        os.unlink(self.filename)
        return report



#   builder.py - Helper classes for building packages
#
#   Copyright (C) 2010, Benjamin Drung <bdrung@ubuntu.com>
#   Copyright (C) 2010, Evan Broder <evan@ebroder.net>
#
#   Permission to use, copy, modify, and/or distribute this software
#   for any purpose with or without fee is hereby granted, provided
#   that the above copyright notice and this permission notice appear
#   in all copies.
#
#   THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
#   WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
#   WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE
#   AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
#   CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
#   LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
#   NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
#   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

import logging
import os
import subprocess

Logger = logging.getLogger(__name__)


def _build_preparation(result_directory):
    """prepares the builder for building a package"""
    if not os.path.isdir(result_directory):
        os.makedirs(result_directory)


class Builder:
    def __init__(self, name):
        self.name = name
        cmd = ["dpkg-architecture", "-qDEB_BUILD_ARCH_CPU"]
        self.architecture = subprocess.check_output(cmd, encoding="utf-8").strip()

    def _build_failure(self, returncode, dsc_file):
        if returncode != 0:
            Logger.error(
                "Failed to build %s from source with %s.", os.path.basename(dsc_file), self.name
            )
        return returncode

    def exists_in_path(self):
        for path in os.environ.get("PATH", os.defpath).split(os.pathsep):
            if os.path.isfile(os.path.join(path, self.name)):
                return True
        return False

    def get_architecture(self):
        return self.architecture

    def get_name(self):
        return self.name

    def _update_failure(self, returncode, dist):
        if returncode != 0:
            Logger.error("Failed to update %s chroot for %s.", dist, self.name)
        return returncode


class Pbuilder(Builder):
    def __init__(self, name="pbuilder"):
        Builder.__init__(self, name)

    def build(self, dsc_file, dist, result_directory):
        _build_preparation(result_directory)
        cmd = [
            "sudo",
            "-E",
            f"ARCH={self.architecture}",
            f"DIST={dist}",
            self.name,
            "--build",
            "--architecture",
            self.architecture,
            "--distribution",
            dist,
            "--buildresult",
            result_directory,
            dsc_file,
        ]
        Logger.debug(" ".join(cmd))
        returncode = subprocess.call(cmd)
        return self._build_failure(returncode, dsc_file)

    def update(self, dist):
        cmd = [
            "sudo",
            "-E",
            f"ARCH={self.architecture}",
            f"DIST={dist}",
            self.name,
            "--update",
            "--architecture",
            self.architecture,
            "--distribution",
            dist,
        ]
        Logger.debug(" ".join(cmd))
        returncode = subprocess.call(cmd)
        return self._update_failure(returncode, dist)


class Pbuilderdist(Builder):
    def __init__(self, name="pbuilder-dist"):
        Builder.__init__(self, name)

    def build(self, dsc_file, dist, result_directory):
        _build_preparation(result_directory)
        cmd = [
            self.name,
            dist,
            self.architecture,
            "build",
            dsc_file,
            "--buildresult",
            result_directory,
        ]
        Logger.debug(" ".join(cmd))
        returncode = subprocess.call(cmd)
        return self._build_failure(returncode, dsc_file)

    def update(self, dist):
        cmd = [self.name, dist, self.architecture, "update"]
        Logger.debug(" ".join(cmd))
        returncode = subprocess.call(cmd)
        return self._update_failure(returncode, dist)


class Sbuild(Builder):
    def __init__(self):
        Builder.__init__(self, "sbuild")

    def build(self, dsc_file, dist, result_directory):
        _build_preparation(result_directory)
        workdir = os.getcwd()
        Logger.debug("cd %s", result_directory)
        os.chdir(result_directory)
        cmd = ["sbuild", "--arch-all", f"--dist={dist}", f"--arch={self.architecture}", dsc_file]
        Logger.debug(" ".join(cmd))
        returncode = subprocess.call(cmd)
        Logger.debug("cd %s", workdir)
        os.chdir(workdir)
        return self._build_failure(returncode, dsc_file)

    def update(self, dist):
        cmd = ["schroot", "--list"]
        Logger.debug(" ".join(cmd))
        process = subprocess.run(cmd, check=False, stdout=subprocess.PIPE, encoding="utf-8")
        chroots, _ = process.stdout.strip().split()
        if process.returncode != 0:
            return process.returncode

        params = {"dist": dist, "arch": self.architecture}
        for chroot in (
            "%(dist)s-%(arch)s-sbuild-source",
            "%(dist)s-sbuild-source",
            "%(dist)s-%(arch)s-source",
            "%(dist)s-source",
        ):
            chroot = chroot % params
            if chroot in chroots:
                break
        else:
            return 1

        commands = [["sbuild-update"], ["sbuild-distupgrade"], ["sbuild-clean", "-a", "-c"]]
        for cmd in commands:
            # pylint: disable=W0631
            Logger.debug("%s %s", " ".join(cmd), chroot)
            ret = subprocess.call(cmd + [chroot])
            # pylint: enable=W0631
            if ret != 0:
                return self._update_failure(ret, dist)
        return 0


_SUPPORTED_BUILDERS = {
    "cowbuilder": lambda: Pbuilder("cowbuilder"),
    "cowbuilder-dist": lambda: Pbuilderdist("cowbuilder-dist"),
    "pbuilder": Pbuilder,
    "pbuilder-dist": Pbuilderdist,
    "sbuild": Sbuild,
}


def get_builder(name):
    if name in _SUPPORTED_BUILDERS:
        builder = _SUPPORTED_BUILDERS[name]()
        if builder.exists_in_path():
            return builder
        Logger.error("Builder doesn't appear to be installed: %s", name)
    else:
        Logger.error("Unsupported builder specified: %s.", name)
        Logger.error("Supported builders: %s", ", ".join(sorted(_SUPPORTED_BUILDERS.keys())))
    return None



# -*- coding: utf-8 -*-
#
# Ubuntu Development Tools
# https://launchpad.net/ubuntu-dev-tools

import logging
import sys


def getLogger():  # pylint: disable=invalid-name
    """Get the logger instance for this module

    Quick guide for using this or not: if you want to call ubuntutools
    module code and have its output print to stdout/stderr ONLY, you can
    use the logger this creates. You can also log directly to this logger
    from your own code to send output to stdout/stderr.

    This creates the ubuntutools module-level logger, and sets some default
    values for formatting and levels, and directs INFO-level logs messages to
    stdout and logs higher than INFO to stderr. The logger's level may be
    adjusted to show more logs (e.g. DEBUG) or less (e.g. WARNING, to suppress
    all INFO messages).

    Without calling this module, the ubuntutools logs will propagate up to
    higher level loggers (possibly the root logger) and be handled by them.
    Note that the default for python logging is to print WARNING and above
    logs to stderr.

    Note if any code calls this method, the ubuntutools module-level logger
    will no longer propagate ubuntutools log message up to higher level
    loggers.

    This should only be used by runnable scripts provided by the
    ubuntu-dev-tools package, or other runnable scripts that want the behavior
    described above.
    """
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)
    logger.propagate = False

    fmt = logging.Formatter("%(message)s")

    stdout_handler = logging.StreamHandler(stream=sys.stdout)
    stdout_handler.setFormatter(fmt)
    stdout_handler.addFilter(lambda r: r.levelno <= logging.INFO)
    logger.addHandler(stdout_handler)

    stderr_handler = logging.StreamHandler(stream=sys.stderr)
    stderr_handler.setFormatter(fmt)
    stderr_handler.setLevel(logging.INFO + 1)
    logger.addHandler(stderr_handler)

    return logger



# test_config.py - Test suite for ubuntutools.config
# -*- coding: utf-8 -*-
#
# Copyright (C) 2010, Stefano Rivera <stefanor@ubuntu.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.

import locale
import os
import unittest
from io import StringIO
from unittest import mock

from ubuntutools.config import UDTConfig, ubu_email


class ConfigTestCase(unittest.TestCase):
    _config_files = {"system": "", "user": ""}

    def _fake_open(self, filename, mode="r", encoding=None):
        self.assertTrue(encoding, f"encoding for {filename} not specified")
        if mode != "r":
            raise IOError("Read only fake-file")
        files = {
            "/etc/devscripts.conf": self._config_files["system"],
            os.path.expanduser("~/.devscripts"): self._config_files["user"],
        }
        if filename not in files:
            raise IOError(f"No such file or directory: '{filename}'")
        return StringIO(files[filename])

    def setUp(self):
        super().setUp()
        open_mock = mock.mock_open()
        open_mock.side_effect = self._fake_open
        patcher = mock.patch("builtins.open", open_mock)
        self.addCleanup(patcher.stop)
        patcher.start()

        # Logger.stdout = StringIO()
        # Logger.stderr = StringIO()

        self.clean_environment()

    def tearDown(self):
        # self.assertEqual(Logger.stdout.getvalue(), '')
        # self.assertEqual(Logger.stderr.getvalue(), '')
        # Logger.stdout = sys.stdout
        # Logger.stderr = sys.stderr

        self.clean_environment()

    def clean_environment(self):
        self._config_files["system"] = ""
        self._config_files["user"] = ""
        for k in list(os.environ.keys()):
            if k.startswith(("UBUNTUTOOLS_", "TEST_")):
                del os.environ[k]

    def test_config_parsing(self):
        self._config_files["user"] = """#COMMENT=yes
\tTAB_INDENTED=yes
 SPACE_INDENTED=yes
SPACE_SUFFIX=yes
SINGLE_QUOTE='yes no'
DOUBLE_QUOTE="yes no"
QUOTED_QUOTE="it's"
PAIR_QUOTES="yes "a' no'
COMMAND_EXECUTION=a b
INHERIT=user
REPEAT=no
REPEAT=yes
"""
        self._config_files["system"] = "INHERIT=system"
        self.assertEqual(
            UDTConfig(prefix="TEST").config,
            {
                "TAB_INDENTED": "yes",
                "SPACE_INDENTED": "yes",
                "SPACE_SUFFIX": "yes",
                "SINGLE_QUOTE": "yes no",
                "DOUBLE_QUOTE": "yes no",
                "QUOTED_QUOTE": "it's",
                "PAIR_QUOTES": "yes a no",
                "COMMAND_EXECUTION": "a",
                "INHERIT": "user",
                "REPEAT": "yes",
            },
        )
        # errs = Logger.stderr.getvalue().strip()
        # Logger.stderr = StringIO()
        # self.assertEqual(len(errs.splitlines()), 1)
        # self.assertRegex(errs,
        # r'Warning: Cannot parse.*\bCOMMAND_EXECUTION=a')

    @staticmethod
    def get_value(*args, **kwargs):
        config = UDTConfig(prefix="TEST")
        return config.get_value(*args, **kwargs)

    def test_defaults(self):
        self.assertEqual(self.get_value("BUILDER"), "pbuilder")

    def test_provided_default(self):
        self.assertEqual(self.get_value("BUILDER", default="foo"), "foo")

    def test_scriptname_precedence(self):
        self._config_files["user"] = "TEST_BUILDER=foo\nUBUNTUTOOLS_BUILDER=bar\n"
        self.assertEqual(self.get_value("BUILDER"), "foo")

    def test_configfile_precedence(self):
        self._config_files["system"] = "UBUNTUTOOLS_BUILDER=foo"
        self._config_files["user"] = "UBUNTUTOOLS_BUILDER=bar"
        self.assertEqual(self.get_value("BUILDER"), "bar")

    def test_environment_precedence(self):
        self._config_files["user"] = "UBUNTUTOOLS_BUILDER=bar"
        os.environ["UBUNTUTOOLS_BUILDER"] = "baz"
        self.assertEqual(self.get_value("BUILDER"), "baz")

    def test_general_environment_specific_config_precedence(self):
        self._config_files["user"] = "TEST_BUILDER=bar"
        os.environ["UBUNTUTOOLS_BUILDER"] = "foo"
        self.assertEqual(self.get_value("BUILDER"), "bar")

    def test_compat_keys(self):
        self._config_files["user"] = "COMPATFOOBAR=bar"
        self.assertEqual(self.get_value("QUX", compat_keys=["COMPATFOOBAR"]), "bar")
        # errs = Logger.stderr.getvalue().strip()
        # Logger.stderr = StringIO()
        # self.assertEqual(len(errs.splitlines()), 1)
        # self.assertRegex(errs,
        # r'deprecated.*\bCOMPATFOOBAR\b.*\bTEST_QUX\b')

    def test_boolean(self):
        self._config_files["user"] = "TEST_BOOLEAN=yes"
        self.assertEqual(self.get_value("BOOLEAN", boolean=True), True)
        self._config_files["user"] = "TEST_BOOLEAN=no"
        self.assertEqual(self.get_value("BOOLEAN", boolean=True), False)
        self._config_files["user"] = "TEST_BOOLEAN=true"
        self.assertEqual(self.get_value("BOOLEAN", boolean=True), None)

    def test_nonpackagewide(self):
        self._config_files["user"] = "UBUNTUTOOLS_FOOBAR=a"
        self.assertEqual(self.get_value("FOOBAR"), None)


class UbuEmailTestCase(unittest.TestCase):
    def setUp(self):
        self.clean_environment()

    def tearDown(self):
        self.clean_environment()

    @staticmethod
    def clean_environment():
        for k in ("UBUMAIL", "DEBEMAIL", "DEBFULLNAME"):
            if k in os.environ:
                del os.environ[k]

    def test_pristine(self):
        os.environ["DEBFULLNAME"] = name = "Joe Developer"
        os.environ["DEBEMAIL"] = email = "joe@example.net"
        self.assertEqual(ubu_email(), (name, email))

    def test_two_hat(self):
        os.environ["DEBFULLNAME"] = name = "Joe Developer"
        os.environ["DEBEMAIL"] = "joe@debian.org"
        os.environ["UBUMAIL"] = email = "joe@ubuntu.com"
        self.assertEqual(ubu_email(), (name, email))
        self.assertEqual(os.environ["DEBFULLNAME"], name)
        self.assertEqual(os.environ["DEBEMAIL"], email)

    def test_two_hat_cmdlineoverride(self):
        os.environ["DEBFULLNAME"] = "Joe Developer"
        os.environ["DEBEMAIL"] = "joe@debian.org"
        os.environ["UBUMAIL"] = "joe@ubuntu.com"
        name = "Foo Bar"
        email = "joe@example.net"
        self.assertEqual(ubu_email(name, email), (name, email))
        self.assertEqual(os.environ["DEBFULLNAME"], name)
        self.assertEqual(os.environ["DEBEMAIL"], email)

    def test_two_hat_noexport(self):
        os.environ["DEBFULLNAME"] = name = "Joe Developer"
        os.environ["DEBEMAIL"] = demail = "joe@debian.org"
        os.environ["UBUMAIL"] = uemail = "joe@ubuntu.com"
        self.assertEqual(ubu_email(export=False), (name, uemail))
        self.assertEqual(os.environ["DEBFULLNAME"], name)
        self.assertEqual(os.environ["DEBEMAIL"], demail)

    def test_two_hat_with_name(self):
        os.environ["DEBFULLNAME"] = "Joe Developer"
        os.environ["DEBEMAIL"] = "joe@debian.org"
        name = "Joe Ubuntunista"
        email = "joe@ubuntu.com"
        os.environ["UBUMAIL"] = f"{name} <{email}>"
        self.assertEqual(ubu_email(), (name, email))
        self.assertEqual(os.environ["DEBFULLNAME"], name)
        self.assertEqual(os.environ["DEBEMAIL"], email)

    def test_debemail_with_name(self):
        name = "Joe Developer"
        email = "joe@example.net"
        os.environ["DEBEMAIL"] = orig = f"{name} <{email}>"
        self.assertEqual(ubu_email(), (name, email))
        self.assertEqual(os.environ["DEBEMAIL"], orig)

    def test_unicode_name(self):
        encoding = locale.getlocale()[1]
        if not encoding:
            encoding = "utf-8"
        name = "Jöe Déveloper"
        env_name = name
        if isinstance(name, bytes):
            name = "Jöe Déveloper".decode("utf-8")
            env_name = name.encode(encoding)
        try:
            os.environ["DEBFULLNAME"] = env_name
        except UnicodeEncodeError:
            self.skipTest("python interpreter is not running in an unicode capable locale")
        os.environ["DEBEMAIL"] = email = "joe@example.net"
        self.assertEqual(ubu_email(), (name, email))



# Copyright (C) 2024 Canonical Ltd.
# Author: Chris Peterson <chris.peterson@canonical.com>
# Author: Andy P. Whitcroft
# Author: Christian Ehrhardt
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import json
import sys
import urllib
import urllib.request

URL_RUNNING = "http://autopkgtest.ubuntu.com/static/running.json"
URL_QUEUED = "http://autopkgtest.ubuntu.com/queues.json"


def _get_jobs(url: str) -> dict:
    request = urllib.request.Request(url, headers={"Cache-Control": "max-age-0"})
    with urllib.request.urlopen(request) as response:
        data = response.read()
        jobs = json.loads(data.decode("utf-8"))

    return jobs


def get_running():
    jobs = _get_jobs(URL_RUNNING)

    running = []
    for pkg in jobs:
        for handle in jobs[pkg]:
            for series in jobs[pkg][handle]:
                for arch in jobs[pkg][handle][series]:
                    jobinfo = jobs[pkg][handle][series][arch]
                    triggers = ",".join(jobinfo[0].get("triggers", "-"))
                    ppas = ",".join(jobinfo[0].get("ppas", "-"))
                    time = jobinfo[1]
                    env = jobinfo[0].get("env", "-")
                    time = str(datetime.timedelta(seconds=jobinfo[1]))
                    try:
                        line = (
                            f"R     {time:6} {pkg:30} {'-':10} {series:8} {arch:8}"
                            f" {ppas:31} {triggers} {env}\n"
                        )
                        running.append((jobinfo[1], line))
                    except BrokenPipeError:
                        sys.exit(1)

    output = ""
    for time, row in sorted(running, reverse=True):
        output += f"{row}"

    return output


def get_queued():
    queues = _get_jobs(URL_QUEUED)
    output = ""
    for origin in queues:
        for series in queues[origin]:
            for arch in queues[origin][series]:
                n = 0
                for key in queues[origin][series][arch]:
                    if key == "private job":
                        pkg = triggers = ppas = "private job"
                    else:
                        pkg, json_data = key.split(maxsplit=1)
                        try:
                            jobinfo = json.loads(json_data)
                            triggers = ",".join(jobinfo.get("triggers", "-"))
                            ppas = ",".join(jobinfo.get("ppas", "-"))
                        except json.decoder.JSONDecodeError:
                            pkg = triggers = ppas = "failed to parse"
                            continue

                    n = n + 1
                    try:
                        output += (
                            f"Q{n:04d} {'-:--':>6} {pkg:30} {origin:10} {series:8} {arch:8}"
                            f" {ppas:31} {triggers}\n"
                        )
                    except BrokenPipeError:
                        sys.exit(1)
    return output



# test_archive.py - Test suite for ubuntutools.archive
#
# Copyright (C) 2010-2012, Stefano Rivera <stefanor@ubuntu.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.


import filecmp
import tempfile
import unittest
from pathlib import Path

import ubuntutools.archive
from ubuntutools.test.example_package import ExamplePackage


class BaseVerificationTestCase(unittest.TestCase):
    def setUp(self):
        tmpdir = tempfile.TemporaryDirectory()
        self.addCleanup(tmpdir.cleanup)
        self.pkg = ExamplePackage(destdir=Path(tmpdir.name))
        self.pkg.create()
        self.dsc = ubuntutools.archive.Dsc(self.pkg.dsc.read_bytes())


class DscVerificationTestCase(BaseVerificationTestCase):
    def test_good(self):
        self.assertTrue(self.dsc.verify_file(self.pkg.orig))
        self.assertTrue(self.dsc.verify_file(self.pkg.debian))

    def test_missing(self):
        self.assertFalse(self.dsc.verify_file(self.pkg.destdir / "does.not.exist"))

    def test_bad(self):
        data = self.pkg.orig.read_bytes()
        last_byte = chr(data[-1] ^ 8).encode()
        data = data[:-1] + last_byte
        self.pkg.orig.write_bytes(data)
        self.assertFalse(self.dsc.verify_file(self.pkg.orig))

    def test_sha1(self):
        del self.dsc["Checksums-Sha256"]
        self.test_good()
        self.test_bad()

    def test_md5(self):
        del self.dsc["Checksums-Sha256"]
        del self.dsc["Checksums-Sha1"]
        self.test_good()
        self.test_bad()


class LocalSourcePackageTestCase(BaseVerificationTestCase):
    SourcePackage = ubuntutools.archive.UbuntuSourcePackage

    def setUp(self):
        super().setUp()
        tmpdir = tempfile.TemporaryDirectory()
        self.addCleanup(tmpdir.cleanup)
        self.workdir = Path(tmpdir.name)

    def pull(self, **kwargs):
        """Do the pull from pkg dir to the workdir, return the SourcePackage"""
        srcpkg = self.SourcePackage(dscfile=self.pkg.dsc, workdir=self.workdir, **kwargs)
        srcpkg.pull()
        return srcpkg

    def _test_pull(self, **kwargs):
        srcpkg = self.pull(**kwargs)
        self.assertTrue(filecmp.cmp(self.pkg.dsc, self.workdir / self.pkg.dsc.name))
        self.assertTrue(filecmp.cmp(self.pkg.orig, self.workdir / self.pkg.orig.name))
        self.assertTrue(filecmp.cmp(self.pkg.debian, self.workdir / self.pkg.debian.name))
        return srcpkg

    def test_unpack(self, **kwargs):
        srcpkg = kwargs.get("srcpkg", self.pull(**kwargs))
        srcpkg.unpack()
        content = self.workdir / self.pkg.dirname / self.pkg.content_filename
        self.assertEqual(self.pkg.content_text, content.read_text())
        debian = self.workdir / self.pkg.dirname / "debian"
        self.assertTrue(debian.exists())
        self.assertTrue(debian.is_dir())

    def test_pull_and_unpack(self, **kwargs):
        self.test_unpack(srcpkg=self._test_pull(**kwargs))

    def test_with_package(self):
        self.test_pull_and_unpack(package=self.pkg.source)

    def test_with_package_version(self):
        self.test_pull_and_unpack(package=self.pkg.source, version=self.pkg.version)

    def test_with_package_version_component(self):
        self.test_pull_and_unpack(
            package=self.pkg.source, version=self.pkg.version, componet="main"
        )

    def test_verification(self):
        corruption = b"CORRUPTION"

        self.pull()

        testfile = self.workdir / self.pkg.debian.name
        self.assertTrue(testfile.exists())
        self.assertTrue(testfile.is_file())
        self.assertNotEqual(testfile.read_bytes(), corruption)
        testfile.write_bytes(corruption)
        self.assertEqual(testfile.read_bytes(), corruption)

        self._test_pull()
        self.assertTrue(testfile.exists())
        self.assertTrue(testfile.is_file())
        self.assertNotEqual(testfile.read_bytes(), corruption)



# Copyright (C) 2019-2023 Canonical Ltd.
# Author: Brian Murray <brian.murray@canonical.com> et al.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Portions of archive related code that is re-used by various tools."""

import os
import re
import urllib.request
from datetime import datetime

import dateutil.parser
from dateutil.tz import tzutc


def get_cache_dir():
    cache_dir = os.environ.get("XDG_CACHE_HOME", os.path.expanduser(os.path.join("~", ".cache")))
    uat_cache = os.path.join(cache_dir, "ubuntu-archive-tools")
    os.makedirs(uat_cache, exist_ok=True)
    return uat_cache


def get_url(url, force_cached):
    """Return file to the URL, possibly caching it"""
    cache_file = None

    # ignore bileto urls wrt caching, they're usually too small to matter
    # and we don't do proper cache expiry
    m = re.search("ubuntu-archive-team.ubuntu.com/proposed-migration/([^/]*)/([^/]*)", url)
    if m:
        cache_dir = get_cache_dir()
        cache_file = os.path.join(cache_dir, f"{m.group(1)}_{m.group(2)}")
    else:
        # test logs can be cached, too
        m = re.search(
            "https://autopkgtest.ubuntu.com/results/autopkgtest-[^/]*/([^/]*)/([^/]*)"
            "/[a-z0-9]*/([^/]*)/([_a-f0-9]*)@/log.gz",
            url,
        )
        if m:
            cache_dir = get_cache_dir()
            cache_file = os.path.join(
                cache_dir, f"{m.group(1)}_{m.group(2)}_{m.group(3)}_{m.group(4)}.gz"
            )

    if cache_file:
        try:
            prev_mtime = os.stat(cache_file).st_mtime
        except FileNotFoundError:
            prev_mtime = 0
        prev_timestamp = datetime.fromtimestamp(prev_mtime, tz=tzutc())
        new_timestamp = datetime.now(tz=tzutc()).timestamp()
        if force_cached:
            return open(cache_file, "rb")

    f = urllib.request.urlopen(url)

    if cache_file:
        remote_ts = dateutil.parser.parse(f.headers["last-modified"])
        if remote_ts > prev_timestamp:
            with open(f"{cache_file}.new", "wb") as new_cache:
                for line in f:
                    new_cache.write(line)
            os.rename(f"{cache_file}.new", cache_file)
            os.utime(cache_file, times=(new_timestamp, new_timestamp))
        f.close()
        f = open(cache_file, "rb")
    return f



# Copyright (C) 2011, Stefano Rivera <stefanor@ubuntu.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import json
import os

import httplib2


class RDependsException(Exception):
    pass


def query_rdepends(package, release, arch, server="http://qa.ubuntuwire.org/rdepends"):
    """Look up a packages reverse-dependencies on the Ubuntuwire
    Reverse- webservice
    """

    url = os.path.join(server, "v1", release, arch, package)

    response, data = httplib2.Http().request(url)
    if response.status != 200:
        raise RDependsException(data.strip())
    return json.loads(data)



#
# misc.py - misc functions for the Ubuntu Developer Tools scripts.
#
# Copyright (C) 2008,      Jonathan Davies <jpds@ubuntu.com>,
#               2008-2009, Siegfried-Angel Gevatter Pujals <rainct@ubuntu.com>,
#               2010,      Stefano Rivera <stefanor@ubuntu.com>
#               2011,      Evan Broder <evan@ebroder.net>
#
# ##################################################################
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# See file /usr/share/common-licenses/GPL for more details.
#
# ##################################################################

import hashlib
import locale
import logging
import os
import shutil
import sys
import tempfile
from contextlib import suppress
from pathlib import Path
from subprocess import CalledProcessError, check_output
from urllib.parse import urlparse

import distro_info
import requests

from ubuntutools.lp.udtexceptions import PocketDoesNotExistError

Logger = logging.getLogger(__name__)


DEFAULT_POCKETS = ("Release", "Security", "Updates", "Proposed")
POCKETS = DEFAULT_POCKETS + ("Backports",)

DEFAULT_STATUSES = ("Pending", "Published")
STATUSES = DEFAULT_STATUSES + ("Superseded", "Deleted", "Obsolete")

UPLOAD_QUEUE_STATUSES = ("New", "Unapproved", "Accepted", "Done", "Rejected")

DOWNLOAD_BLOCKSIZE_DEFAULT = 8192

_SYSTEM_DISTRIBUTION_CHAIN: list[str] = []


class DownloadError(Exception):
    "Unable to pull a source package"


class NotFoundError(DownloadError):
    "Source package not found"


def system_distribution_chain():
    """system_distribution_chain() -> [string]

    Detect the system's distribution as well as all of its parent
    distributions and return them as a list of strings, with the
    system distribution first (and the greatest grandparent last). If
    the distribution chain can't be determined, print an error message
    and return an empty list.
    """
    if len(_SYSTEM_DISTRIBUTION_CHAIN) == 0:
        try:
            vendor = check_output(("dpkg-vendor", "--query", "Vendor"), encoding="utf-8").strip()
            _SYSTEM_DISTRIBUTION_CHAIN.append(vendor)
        except CalledProcessError:
            Logger.error("Could not determine what distribution you are running.")
            return []

        while True:
            try:
                parent = check_output(
                    (
                        "dpkg-vendor",
                        "--vendor",
                        _SYSTEM_DISTRIBUTION_CHAIN[-1],
                        "--query",
                        "Parent",
                    ),
                    encoding="utf-8",
                ).strip()
            except CalledProcessError:
                # Vendor has no parent
                break
            _SYSTEM_DISTRIBUTION_CHAIN.append(parent)

    return _SYSTEM_DISTRIBUTION_CHAIN


def system_distribution():
    """system_distro() -> string

    Detect the system's distribution and return it as a string. If the
    name of the distribution can't be determined, print an error message
    and return None.
    """
    return system_distribution_chain()[0]


def host_architecture():
    """host_architecture -> string

    Detect the host's architecture and return it as a string. If the
    architecture can't be determined, print an error message and return None.
    """

    try:
        arch = check_output(("dpkg", "--print-architecture"), encoding="utf-8").strip()
    except CalledProcessError:
        arch = None

    if not arch or "not found" in arch:
        Logger.error("Not running on a Debian based system; could not detect its architecture.")
        return None

    return arch


def readlist(filename, uniq=True):
    """readlist(filename, uniq) -> list

    Read a list of words from the indicated file. If 'uniq' is True, filter
    out duplicated words.
    """
    path = Path(filename)

    if not path.is_file():
        Logger.error("File %s does not exist.", path)
        return False

    content = path.read_text(encoding="utf-8").replace("\n", " ").replace(",", " ")

    if not content.strip():
        Logger.error("File %s is empty.", path)
        return False

    items = [item for item in content.split() if item]

    if uniq:
        items = list(set(items))

    return items


def split_release_pocket(release, default="Release"):
    """Splits the release and pocket name.

    If the argument doesn't contain a pocket name then the 'Release' pocket
    is assumed.

    Returns the release and pocket name.
    """
    pocket = default

    if release is None:
        raise ValueError("No release name specified")

    if "-" in release:
        release, pocket = release.rsplit("-", 1)
        pocket = pocket.capitalize()

        if pocket not in POCKETS:
            raise PocketDoesNotExistError(f"Pocket '{pocket}' does not exist.")

    return (release, pocket)


def require_utf8():
    """Can be called by programs that only function in UTF-8 locales"""
    if locale.getpreferredencoding() != "UTF-8":
        Logger.error("This program only functions in a UTF-8 locale. Aborting.")
        sys.exit(1)


_vendor_to_distroinfo = {
    "Debian": distro_info.DebianDistroInfo,
    "Ubuntu": distro_info.UbuntuDistroInfo,
}


def vendor_to_distroinfo(vendor):
    """vendor_to_distroinfo(string) -> DistroInfo class

    Convert a string name of a distribution into a DistroInfo subclass
    representing that distribution, or None if the distribution is
    unknown.
    """
    return _vendor_to_distroinfo.get(vendor)


def codename_to_distribution(codename):
    """codename_to_distribution(string) -> string

    Finds a given release codename in your distribution's genaology
    (i.e. looking at the current distribution and its parents), or
    print an error message and return None if it can't be found
    """
    for distro in system_distribution_chain() + ["Ubuntu", "Debian"]:
        info = vendor_to_distroinfo(distro)
        if not info:
            continue

        if info().valid(codename):
            return distro
    return None


def verify_file_checksums(pathname, checksums=None, size=0):
    """verify checksums of file

    Any failure will log an error.

    pathname: str or Path
        full path to file
    checksums: dict
        keys are alg name, values are expected checksum
    size: int
        size of file, if known

    Returns True if all checks pass, False otherwise
    """
    if checksums is None:
        checksums = {}
    path = Path(pathname)

    if not path.is_file():
        Logger.error("File %s not found", path)
        return False
    filesize = path.stat().st_size
    if size and size != filesize:
        Logger.error("File %s incorrect size, got %s expected %s", path, filesize, size)
        return False

    for alg, checksum in checksums.items():
        hash_ = hashlib.new(alg)
        with path.open("rb") as f:
            while True:
                block = f.read(hash_.block_size)
                if len(block) == 0:
                    break
                hash_.update(block)
        digest = hash_.hexdigest()
        if digest == checksum:
            Logger.debug("File %s checksum (%s) verified: %s", path, alg, checksum)
        else:
            Logger.error(
                "File %s checksum (%s) mismatch: got %s expected %s", path, alg, digest, checksum
            )
            return False
    return True


def verify_file_checksum(pathname, alg, checksum, size=0):
    """verify checksum of file

    pathname: str or Path
        full path to file
    alg: str
        name of checksum alg
    checksum: str
        expected checksum
    size: int
        size of file, if known

    Returns True if all checks pass, False otherwise
    """
    return verify_file_checksums(pathname, {alg: checksum}, size)


def extract_authentication(url):
    """Remove plaintext authentication data from a URL

    If the URL has a username:password in its netloc, this removes it
    and returns the remaining URL, along with the username and password
    separately. If no authentication data is in the netloc, this just
    returns the URL unchanged with None for the username and password.

    This returns a tuple in the form (url, username, password)
    """
    components = urlparse(url)
    if components.username or components.password:
        return (
            components._replace(netloc=components.hostname).geturl(),
            components.username,
            components.password,
        )
    return (url, None, None)


def download(src, dst, size=0, *, blocksize=DOWNLOAD_BLOCKSIZE_DEFAULT):
    """download/copy a file/url to local file

    src: str or Path
        Source to copy from (file path or url)
    dst: str or Path
        Destination dir or filename
    size: int
        Size of source, if known
    blocksize: int or None
        Blocksize to use when downloading

    If the URL contains authentication data in the URL 'netloc',
    it will be stripped from the URL and passed to the requests library.

    This may throw a DownloadError.

    On success, this will return the dst as a Path object.
    """
    src = str(src)
    parsedsrc = urlparse(src)

    dst = Path(dst).expanduser().resolve()
    if dst.is_dir():
        dst = dst / Path(parsedsrc.path).name

    # Copy if src is a local file
    if parsedsrc.scheme in ["", "file"]:
        src = Path(parsedsrc.path).expanduser().resolve()
        if src != parsedsrc.path:
            Logger.info("Parsed %s as %s", parsedsrc.path, src)
        if not src.exists():
            raise NotFoundError(f"Source file {src} not found")
        if dst.exists():
            if src.samefile(dst):
                Logger.info("Using existing file %s", dst)
                return dst
            Logger.info("Replacing existing file %s", dst)
        Logger.info("Copying file %s to %s", src, dst)
        shutil.copyfile(src, dst)
        return dst

    src, username, password = extract_authentication(src)
    auth = (username, password) if username or password else None

    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdst = Path(tmpdir) / "dst"
        try:
            # We must use "Accept-Encoding: identity" so that Launchpad doesn't
            # compress changes files. See LP: #2025748.
            with requests.get(
                src, stream=True, timeout=60, auth=auth, headers={"accept-encoding": "identity"}
            ) as fsrc:
                with tmpdst.open("wb") as fdst:
                    fsrc.raise_for_status()
                    _download(fsrc, fdst, size, blocksize=blocksize)
        except requests.exceptions.HTTPError as error:
            if error.response is not None and error.response.status_code == 404:
                raise NotFoundError(f"URL {src} not found: {error}") from error
            raise DownloadError(error) from error
        except requests.exceptions.ConnectionError as error:
            # This is likely a archive hostname that doesn't resolve, like 'ftpmaster.internal'
            raise NotFoundError(f"URL {src} not found: {error}") from error
        except requests.exceptions.RequestException as error:
            raise DownloadError(error) from error
        shutil.move(tmpdst, dst)
    return dst


class _StderrProgressBar:
    BAR_WIDTH_MIN = 40
    BAR_WIDTH_DEFAULT = 60

    def __init__(self, max_width):
        self.full_width = min(max_width, self.BAR_WIDTH_DEFAULT)
        self.width = self.full_width - len("[] 99%")
        self.show_progress = self.full_width >= self.BAR_WIDTH_MIN

    def update(self, progress, total):
        if not self.show_progress:
            return
        pct = progress * 100 // total
        pctstr = f"{pct:>3}%"
        barlen = self.width * pct // 100
        barstr = "=" * barlen
        barstr = f"{barstr[:-1]}>"
        barstr = barstr.ljust(self.width)
        fullstr = f"\r[{barstr}]{pctstr}"
        sys.stderr.write(fullstr)
        sys.stderr.flush()

    def finish(self):
        if not self.show_progress:
            return
        sys.stderr.write("\n")
        sys.stderr.flush()


def _download(fsrc, fdst, size, *, blocksize):
    """helper method to download src to dst using requests library."""
    url = fsrc.url
    Logger.debug("Using URL: %s", url)

    if not size:
        with suppress(AttributeError, TypeError, ValueError):
            size = int(fsrc.headers.get("Content-Length"))

    parsed = urlparse(url)
    filename = Path(parsed.path).name
    hostname = parsed.hostname
    sizemb = f" ({size / 1024.0 / 1024:0.3f} MiB)" if size else ""
    Logger.info("Downloading %s from %s%s", filename, hostname, sizemb)

    # Don't show progress if:
    #   logging INFO is suppressed
    #   stderr isn't a tty
    #   we don't know the total file size
    #   the file is content-encoded (i.e. compressed)
    show_progress = all(
        (
            Logger.isEnabledFor(logging.INFO),
            sys.stderr.isatty(),
            size > 0,
            "Content-Encoding" not in fsrc.headers,
        )
    )

    terminal_width = 0
    if show_progress:
        try:
            terminal_width = os.get_terminal_size(sys.stderr.fileno()).columns
        except Exception as e:  # pylint: disable=broad-except
            Logger.error("Error finding stderr width, suppressing progress bar: %s", e)
    progress_bar = _StderrProgressBar(max_width=terminal_width)

    downloaded = 0
    try:
        while True:
            # We use fsrc.raw so that compressed files stay compressed as we
            # write them to disk. For example, if this is a .diff.gz, then it
            # needs to remain compressed and unmodified to remain valid as part
            # of a source package later, even though Launchpad sends
            # "Content-Encoding: gzip" and the requests library therefore would
            # want to decompress it. See LP: #2025748.
            block = fsrc.raw.read(blocksize)
            if not block:
                break
            fdst.write(block)
            downloaded += len(block)
            progress_bar.update(downloaded, size)
    finally:
        progress_bar.finish()
        if size and size > downloaded:
            Logger.error(
                "Partial download: %0.3f MiB of %0.3f MiB",
                downloaded / 1024.0 / 1024,
                size / 1024.0 / 1024,
            )


def _download_text(src, binary, *, blocksize):
    with tempfile.TemporaryDirectory() as tmpdir:
        dst = Path(tmpdir) / "dst"
        download(src, dst, blocksize=blocksize)
        return dst.read_bytes() if binary else dst.read_text()


def download_text(src, mode=None, *, blocksize=DOWNLOAD_BLOCKSIZE_DEFAULT):
    """Return the text content of a downloaded file

    src: str or Path
        Source to copy from (file path or url)
    mode: str
        Deprecated, ignored unless a string that contains 'b'
    blocksize: int or None
        Blocksize to use when downloading

    Raises the same exceptions as download()

    Returns text content of downloaded file
    """
    return _download_text(src, binary="b" in (mode or ""), blocksize=blocksize)


def download_bytes(src, *, blocksize=DOWNLOAD_BLOCKSIZE_DEFAULT):
    """Same as download_text() but returns bytes"""
    return _download_text(src, binary=True, blocksize=blocksize)



# Copyright (C) 2014, Benjamin Drung <bdrung@ubuntu.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import debian.debian_support


class Version(debian.debian_support.Version):
    def strip_epoch(self):
        """Removes the epoch from a Debian version string.

        strip_epoch(1:1.52-1) will return "1.52-1" and strip_epoch(1.1.3-1)
        will return "1.1.3-1".
        """
        parts = self.full_version.split(":")
        if len(parts) > 1:
            del parts[0]
        version_without_epoch = ":".join(parts)
        return version_without_epoch

    def get_related_debian_version(self):
        """Strip the ubuntu-specific bits off the version"""
        related_debian_version = self.full_version
        uidx = related_debian_version.find("ubuntu")
        if uidx > 0:
            related_debian_version = related_debian_version[:uidx]
        uidx = related_debian_version.find("build")
        if uidx > 0:
            related_debian_version = related_debian_version[:uidx]
        return Version(related_debian_version)

    def is_modified_in_ubuntu(self):
        """Did Ubuntu modify this (and mark the version appropriately)?"""
        return "ubuntu" in self.full_version



#
# question.py - Internal helper class for sponsor-patch
#
# Copyright (C) 2011, Benjamin Drung <bdrung@ubuntu.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import sys
from typing import NoReturn

from ubuntutools.question import Question, YesNoQuestion


def ask_for_ignoring_or_fixing():
    """Ask the user to resolve an issue manually or ignore it.

    Returns false if the user want to fix the issue and returns true if the user
    want to ignore the issue.
    """

    question = Question(["yes", "ignore", "no"])
    answer = question.ask("Do you want to resolve this issue manually", "yes")
    if answer == "no":
        user_abort()
    return answer == "ignore"


def ask_for_manual_fixing():
    """Ask the user to resolve an issue manually."""

    answer = YesNoQuestion().ask("Do you want to resolve this issue manually", "yes")
    if answer == "no":
        user_abort()


def user_abort() -> NoReturn:
    """Print abort and quit the program."""

    print("User abort.")
    sys.exit(2)



# -*- coding: utf-8 -*-
#
#   mail.py - methods used by requestsync when used in "mail" mode
#
#   Copyright © 2009 Michael Bienia <geser@ubuntu.com>,
#               2011 Stefano Rivera <stefanor@ubuntu.com>
#
#   This module may contain code written by other authors/contributors to
#   the main requestsync script. See there for their names.
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; version 2
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   Please see the /usr/share/common-licenses/GPL-2 file for the full text
#   of the GNU General Public License license.

import logging
import os
import re
import smtplib
import socket
import subprocess
import sys
import tempfile

from debian.changelog import Changelog
from distro_info import DebianDistroInfo, DistroDataOutdated

from ubuntutools.archive import DebianSourcePackage, UbuntuSourcePackage
from ubuntutools.lp.udtexceptions import PackageNotFoundException
from ubuntutools.question import YesNoQuestion, confirmation_prompt

Logger = logging.getLogger(__name__)


__all__ = [
    "get_debian_srcpkg",
    "get_ubuntu_srcpkg",
    "need_sponsorship",
    "check_existing_reports",
    "get_ubuntu_delta_changelog",
    "mail_bug",
]


def get_debian_srcpkg(name, release):
    # Canonicalise release:
    debian_info = DebianDistroInfo()
    try:
        codename = debian_info.codename(release, default=release)
        return DebianSourcePackage(package=name, series=codename).lp_spph
    except DistroDataOutdated as e:
        Logger.warning(e)
    except PackageNotFoundException:
        pass
    return DebianSourcePackage(package=name, series=release).lp_spph


def get_ubuntu_srcpkg(name, release, pocket="Proposed"):
    srcpkg = UbuntuSourcePackage(package=name, series=release, pocket=pocket)
    try:
        return srcpkg.lp_spph
    except PackageNotFoundException:
        if pocket != "Release":
            parent_pocket = "Release"
            if pocket == "Updates":
                parent_pocket = "Proposed"
            return get_ubuntu_srcpkg(name, release, parent_pocket)
        raise


def need_sponsorship(name, component, release):
    """
    Ask the user if he has upload permissions for the package or the
    component.
    """

    val = YesNoQuestion().ask(
        f"Do you have upload permissions for the '{component}' component or "
        f"the package '{name}' in Ubuntu {release}?\nIf in doubt answer 'n'.",
        "no",
    )
    return val == "no"


def check_existing_reports(srcpkg):
    """
    Point the user to the URL to manually check for duplicate bug reports.
    """
    print(
        f"Please check on https://bugs.launchpad.net/ubuntu/+source/{srcpkg}/+bugs\n"
        f"for duplicate sync requests before continuing."
    )
    confirmation_prompt()


def get_ubuntu_delta_changelog(srcpkg):
    """
    Download the Ubuntu changelog and extract the entries since the last sync
    from Debian.
    """
    changelog = Changelog(srcpkg.getChangelog())
    if changelog is None:
        return ""
    delta = []
    debian_info = DebianDistroInfo()
    for block in changelog:
        distribution = block.distributions.split()[0].split("-")[0]
        if debian_info.valid(distribution):
            break
        delta += [str(change) for change in block.changes() if change.strip()]

    return "\n".join(delta)


def mail_bug(
    srcpkg,
    subscribe,
    status,
    bugtitle,
    bugtext,
    bug_mail_domain,
    keyid,
    myemailaddr,
    mailserver_host,
    mailserver_port,
    mailserver_user,
    mailserver_pass,
):
    """
    Submit the sync request per email.
    """

    to = f"new@{bug_mail_domain}"

    # generate mailbody
    if srcpkg:
        mailbody = f" affects ubuntu/{srcpkg}\n"
    else:
        mailbody = " affects ubuntu\n"
    mailbody += f"""\
 status {status}
 importance wishlist
 subscribe {subscribe}
 done

{bugtext}"""

    # prepare sign command
    gpg_command = None
    for cmd in ("gnome-gpg", "gpg2", "gpg"):
        if os.access(f"/usr/bin/{cmd}", os.X_OK):
            gpg_command = [cmd]
            break

    if not gpg_command:
        Logger.error("Cannot locate gpg, please install the 'gnupg' package!")
        sys.exit(1)

    gpg_command.append("--clearsign")
    if keyid:
        gpg_command.extend(("-u", keyid))

    # sign the mail body
    gpg = subprocess.Popen(
        gpg_command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, encoding="utf-8"
    )
    signed_report = gpg.communicate(mailbody)[0]
    if gpg.returncode != 0:
        Logger.error("%s failed.", gpg_command[0])
        sys.exit(1)

    # generate email
    mail = f"""\
From: {myemailaddr}
To: {to}
Subject: {bugtitle}
Content-Type: text/plain; charset=UTF-8

{signed_report}"""

    print(f"The final report is:\n{mail}")
    confirmation_prompt()

    # save mail in temporary file
    backup = tempfile.NamedTemporaryFile(
        mode="w",
        delete=False,
        prefix=f"requestsync-{re.sub('[^a-zA-Z0-9_-]', '', bugtitle.replace(' ', '_'))}",
    )
    with backup:
        backup.write(mail)

    Logger.info(
        "The e-mail has been saved in %s and will be deleted after succesful transmission",
        backup.name,
    )

    # connect to the server
    while True:
        try:
            Logger.info("Connecting to %s:%s ...", mailserver_host, mailserver_port)
            smtp = smtplib.SMTP(mailserver_host, mailserver_port)
            break
        except smtplib.SMTPConnectError as error:
            try:
                # py2 path
                # pylint: disable=unsubscriptable-object
                Logger.error(
                    "Could not connect to %s:%s: %s (%i)",
                    mailserver_host,
                    mailserver_port,
                    error[1],
                    error[0],
                )
            except TypeError:
                # pylint: disable=no-member
                Logger.error(
                    "Could not connect to %s:%s: %s (%i)",
                    mailserver_host,
                    mailserver_port,
                    error.strerror,
                    error.errno,
                )
            if error.smtp_code == 421:
                confirmation_prompt(
                    message="This is a temporary error, press [Enter] "
                    "to retry. Press [Ctrl-C] to abort now."
                )
        except socket.error as error:
            try:
                # py2 path
                # pylint: disable=unsubscriptable-object
                Logger.error(
                    "Could not connect to %s:%s: %s (%i)",
                    mailserver_host,
                    mailserver_port,
                    error[1],
                    error[0],
                )
            except TypeError:
                # pylint: disable=no-member
                Logger.error(
                    "Could not connect to %s:%s: %s (%i)",
                    mailserver_host,
                    mailserver_port,
                    error.strerror,
                    error.errno,
                )
            return

    if mailserver_user and mailserver_pass:
        try:
            smtp.login(mailserver_user, mailserver_pass)
        except smtplib.SMTPAuthenticationError:
            Logger.error("Error authenticating to the server: invalid username and password.")
            smtp.quit()
            return
        except smtplib.SMTPException:
            Logger.error("Unknown SMTP error.")
            smtp.quit()
            return

    while True:
        try:
            smtp.sendmail(myemailaddr, to, mail.encode("utf-8"))
            smtp.quit()
            os.remove(backup.name)
            Logger.info("Sync request mailed.")
            break
        except smtplib.SMTPRecipientsRefused as smtperror:
            smtp_code, smtp_message = smtperror.recipients[to]
            Logger.error("Error while sending: %i, %s", smtp_code, smtp_message)
            if smtp_code == 450:
                confirmation_prompt(
                    message="This is a temporary error, press [Enter] "
                    "to retry. Press [Ctrl-C] to abort now."
                )
            else:
                return
        except smtplib.SMTPResponseException as error:
            Logger.error("Error while sending: %i, %s", error.smtp_code, error.smtp_error)
            return
        except smtplib.SMTPServerDisconnected:
            Logger.error("Server disconnected while sending the mail.")
            return



# test_help.py - Ensure scripts can run --help.
#
# Copyright (C) 2010, Stefano Rivera <stefanor@ubuntu.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.

import subprocess
import unittest

from setup import scripts

TIMEOUT = 10


class HelpTestCase(unittest.TestCase):
    def test_script(self):
        for script in scripts:
            with self.subTest(script=script):
                result = subprocess.run(
                    [f"./{script}", "--help"],
                    encoding="UTF-8",
                    timeout=10,
                    check=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
                self.assertFalse(result.stderr.strip())



#!/usr/bin/python3

import glob
import pathlib
import re
from collections.abc import Sequence

from setuptools import setup


def get_debian_version() -> str:
    """Look what Debian version we have."""
    changelog = pathlib.Path(__file__).parent / "debian" / "changelog"
    with changelog.open("r", encoding="utf-8") as changelog_f:
        head = changelog_f.readline()
    match = re.compile(r".*\((.*)\).*").match(head)
    if not match:
        raise ValueError(f"Failed to extract Debian version from '{head}'.")
    return match.group(1)


def make_pep440_compliant(version: str) -> str:
    """Convert the version into a PEP440 compliant version."""
    public_version_re = re.compile(r"^([0-9][0-9.]*(?:(?:a|b|rc|.post|.dev)[0-9]+)*)\+?")
    _, public, local = public_version_re.split(version, maxsplit=1)
    if not local:
        return version
    sanitized_local = re.sub("[+~]+", ".", local).strip(".")
    pep440_version = f"{public}+{sanitized_local}"
    assert re.match("^[a-zA-Z0-9.]+$", sanitized_local), f"'{pep440_version}' not PEP440 compliant"
    return pep440_version


scripts = [
    "backportpackage",
    "check-mir",
    "check-symbols",
    "dch-repeat",
    "grab-merge",
    "grep-merges",
    "import-bug-from-debian",
    "lp-bitesize",
    "merge-changelog",
    "mk-sbuild",
    "pbuilder-dist",
    "pbuilder-dist-simple",
    "pm-helper",
    "pull-pkg",
    "pull-debian-debdiff",
    "pull-debian-source",
    "pull-debian-debs",
    "pull-debian-ddebs",
    "pull-debian-udebs",
    "pull-lp-source",
    "pull-lp-debs",
    "pull-lp-ddebs",
    "pull-lp-udebs",
    "pull-ppa-source",
    "pull-ppa-debs",
    "pull-ppa-ddebs",
    "pull-ppa-udebs",
    "pull-uca-source",
    "pull-uca-debs",
    "pull-uca-ddebs",
    "pull-uca-udebs",
    "requestbackport",
    "requestsync",
    "reverse-depends",
    "running-autopkgtests",
    "seeded-in-ubuntu",
    "setup-packaging-environment",
    "sponsor-patch",
    "submittodebian",
    "syncpackage",
    "ubuntu-build",
    "ubuntu-iso",
    "ubuntu-upload-permission",
    "update-maintainer",
]
data_files: list[tuple[str, Sequence[str]]] = [
    ("share/bash-completion/completions", glob.glob("bash_completion/*")),
    ("share/man/man1", glob.glob("doc/*.1")),
    ("share/man/man5", glob.glob("doc/*.5")),
    ("share/ubuntu-dev-tools", ["enforced-editing-wrapper"]),
]

if __name__ == "__main__":
    setup(
        name="ubuntu-dev-tools",
        version=make_pep440_compliant(get_debian_version()),
        scripts=scripts,
        packages=[
            "ubuntutools",
            "ubuntutools/lp",
            "ubuntutools/requestsync",
            "ubuntutools/sponsor_patch",
            "ubuntutools/test",
        ],
        data_files=data_files,
        test_suite="ubuntutools.test",
    )



# Copyright (C) 2024 Canonical Ltd.
# Author: Chris Peterson <chris.peterson@canonical.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.
"""Tests for running_autopkgtests
Tests using cached data from autopkgtest servers.

These tests only ensure code changes don't change parsing behavior
of the response data. If the response format changes, then the cached
responses will need to change as well.
"""

import unittest
from unittest.mock import patch

from ubuntutools.running_autopkgtests import (
    URL_QUEUED,
    URL_RUNNING,
    _get_jobs,
    get_queued,
    get_running,
)

# Cached binary response data from autopkgtest server
RUN_DATA = (
    b'{"pyatem": {'
    b" \"submit-time_2024-01-19 19:37:36;triggers_['python3-defaults/3.12.1-0ubuntu1'];\":"
    b' {"noble": {"arm64": [{"triggers": ["python3-defaults/3.12.1-0ubuntu1"],'
    b' "submit-time": "2024-01-19 19:37:36"}, 380, "<omitted log>"]}}}}'
)
QUEUED_DATA = (
    b'{"ubuntu": {"noble": {"arm64": ["libobject-accessor-perl {\\"requester\\": \\"someone\\",'
    b' \\"submit-time\\": \\"2024-01-18 01:08:55\\",'
    b' \\"triggers\\": [\\"perl/5.38.2-3\\", \\"liblocale-gettext-perl/1.07-6build1\\"]}"]}}}'
)

# Expected result(s) of parsing the above JSON data
RUNNING_JOB = {
    "pyatem": {
        "submit-time_2024-01-19 19:37:36;triggers_['python3-defaults/3.12.1-0ubuntu1'];": {
            "noble": {
                "arm64": [
                    {
                        "triggers": ["python3-defaults/3.12.1-0ubuntu1"],
                        "submit-time": "2024-01-19 19:37:36",
                    },
                    380,
                    "<omitted log>",
                ]
            }
        }
    }
}

QUEUED_JOB = {
    "ubuntu": {
        "noble": {
            "arm64": [
                'libobject-accessor-perl {"requester": "someone",'
                ' "submit-time": "2024-01-18 01:08:55",'
                ' "triggers": ["perl/5.38.2-3", "liblocale-gettext-perl/1.07-6build1"]}'
            ]
        }
    }
}


PRIVATE_JOB = {"ppa": {"noble": {"arm64": ["private job"]}}}


# Expected textual output of the program based on the above data
RUNNING_OUTPUT = (
    "R     0:06:20 pyatem                         -          noble    arm64"
    "    -                               python3-defaults/3.12.1-0ubuntu1 -\n"
)
QUEUED_OUTPUT = (
    "Q0001   -:-- libobject-accessor-perl        ubuntu     noble    arm64"
    "    -                               perl/5.38.2-3,liblocale-gettext-perl/1.07-6build1\n"
)
PRIVATE_OUTPUT = (
    "Q0001   -:-- private job                    ppa        noble    arm64"
    "    private job                     private job\n"
)


class RunningAutopkgtestTestCase(unittest.TestCase):
    """Assert helper functions parse data correctly"""

    maxDiff = None

    @patch("urllib.request.urlopen")
    def test_get_running_jobs(self, mock_response):
        """Test: Correctly parse autopkgtest json data for running tests"""
        mock_response.return_value.__enter__.return_value.read.return_value = RUN_DATA
        jobs = _get_jobs(URL_RUNNING)
        self.assertEqual(RUNNING_JOB, jobs)

    @patch("urllib.request.urlopen")
    def test_get_queued_jobs(self, mock_response):
        """Test: Correctly parse autopkgtest json data for queued tests"""
        mock_response.return_value.__enter__.return_value.read.return_value = QUEUED_DATA
        jobs = _get_jobs(URL_QUEUED)
        self.assertEqual(QUEUED_JOB, jobs)

    def test_get_running_output(self):
        """Test: Correctly print running tests"""
        with patch("ubuntutools.running_autopkgtests._get_jobs", return_value=RUNNING_JOB):
            self.assertEqual(get_running(), RUNNING_OUTPUT)

    def test_get_queued_output(self):
        """Test: Correctly print queued tests"""
        with patch("ubuntutools.running_autopkgtests._get_jobs", return_value=QUEUED_JOB):
            self.assertEqual(get_queued(), QUEUED_OUTPUT)

    def test_private_queued_job(self):
        """Test: Correctly print queued private job"""
        with patch("ubuntutools.running_autopkgtests._get_jobs", return_value=PRIVATE_JOB):
            self.assertEqual(get_queued(), PRIVATE_OUTPUT)



# test_reverse_depends.py - Test suite for reverse-depends
#
# Copyright (C) 2026, Nadzeya Hutsko <nadzeya.hutsko@canonical.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.

"""Test suite for the reverse-depends script"""

import importlib.machinery
import importlib.util
import os
import unittest
from unittest import mock


def _load_reverse_depends():
    script_path = os.path.join(os.path.dirname(__file__), "..", "..", "reverse-depends")
    script_path = os.path.abspath(script_path)
    loader = importlib.machinery.SourceFileLoader("reverse_depends", script_path)
    spec = importlib.util.spec_from_loader("reverse_depends", loader)
    module = importlib.util.module_from_spec(spec)
    loader.exec_module(module)
    return module


reverse_depends = _load_reverse_depends()


class DisplayVerboseTestCase(unittest.TestCase):
    """Tests for display_verbose output when no reverse dependencies are found"""

    def test_no_results_at_all(self):
        """Empty values dict prints 'No reverse dependencies found'"""
        with mock.patch.object(reverse_depends.Logger, "info") as mock_info:
            reverse_depends.display_verbose("some-package", {})
        mock_info.assert_called_once_with("No reverse dependencies found")

    def test_results_filtered_out(self):
        """Package present in values but with empty data prints filtered message"""
        # Simulate the case where rdeps exist but are all filtered out (e.g. by
        # -R), so build_results populates `result[package] = {}` when all fields
        # are filtered
        values = {"r-cran-bdgraph": {}}
        with mock.patch.object(reverse_depends.Logger, "info") as mock_info:
            reverse_depends.display_verbose("r-cran-bdgraph", values)
        mock_info.assert_called_once_with("No reverse dependencies found with the current filters")

    def test_results_found(self):
        """Non-empty results are displayed without the 'not found' messages"""
        values = {
            "r-cran-bdgraph": {
                "Reverse-Depends": [{"Package": "r-cran-qgraph", "Architectures": ["amd64"]}]
            }
        }
        with mock.patch.object(reverse_depends.Logger, "info") as mock_info:
            reverse_depends.display_verbose("r-cran-bdgraph", values)
        calls = [str(c) for c in mock_info.call_args_list]
        self.assertFalse(
            any("No reverse dependencies found" in c for c in calls),
            "Should not print 'no rdeps' message when results exist",
        )


if __name__ == "__main__":
    unittest.main()



# config.py - Common configuration file and environment variable handling for
#             the ubuntu-dev-tools package.
#
# Copyright (C) 2010, Stefano Rivera <stefanor@ubuntu.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.

import locale
import logging
import os
import pwd
import re
import shlex
import socket
import sys

Logger = logging.getLogger(__name__)


class UDTConfig:
    """Ubuntu Dev Tools configuration file (devscripts config file) and
    environment variable parsing.
    """

    no_conf = False
    # Package wide configuration variables.
    # These are reqired to be used by at least two scripts.
    defaults = {
        "BUILDER": "pbuilder",
        "DEBIAN_MIRROR": "http://deb.debian.org/debian",
        "DEBSEC_MIRROR": "http://security.debian.org",
        "DEBIAN_DDEBS_MIRROR": "http://debug.mirrors.debian.org/debian-debug",
        "LPINSTANCE": "production",
        "MIRROR_FALLBACK": True,
        "UBUNTU_MIRROR": "http://archive.ubuntu.com/ubuntu",
        "UBUNTU_PORTS_MIRROR": "http://ports.ubuntu.com",
        "UBUNTU_DDEBS_MIRROR": "http://ddebs.ubuntu.com",
        "UPDATE_BUILDER": False,
        "WORKDIR": None,
        "KEYID": None,
    }
    # Populated from the configuration files:
    config: dict[str, str] = {}

    def __init__(self, no_conf=False, prefix=None):
        self.no_conf = no_conf
        if prefix is None:
            prefix = os.path.basename(sys.argv[0]).upper().replace("-", "_")
        self.prefix = prefix
        if not no_conf:
            self.config = self.parse_devscripts_config()

    @staticmethod
    def parse_devscripts_config() -> dict[str, str]:
        """Read the devscripts configuration files, and return the values as a
        dictionary
        """
        config = {}
        for filename in ("/etc/devscripts.conf", "~/.devscripts"):
            try:
                with open(os.path.expanduser(filename), "r", encoding="utf-8") as f:
                    content = f.read()
            except IOError:
                continue
            try:
                tokens = shlex.split(content, comments=True)
            except ValueError as e:
                Logger.error("Error parsing %s: %s", filename, e)
                continue
            for token in tokens:
                if "=" in token:
                    key, value = token.split("=", 1)
                    config[key] = value
        return config

    def get_value(self, key, default=None, boolean=False, compat_keys=()):
        """Retrieve a value from the environment or configuration files.
        keys are prefixed with the script name, falling back to UBUNTUTOOLS for
        package-wide keys.

        Variable Priority: PREFIX_KEY, UBUNTUTOOLS_KEY, compat_keys
        Store Priority: Environment variables, user conf, system conf

        Historical variable names can be supplied via compat_keys, no prefix is
        applied to them.
        """
        if default is None and key in self.defaults:
            default = self.defaults[key]

        keys = [f"{self.prefix}_{key}"]
        if key in self.defaults:
            keys.append(f"UBUNTUTOOLS_{key}")
        keys += compat_keys

        for k in keys:
            for store in (os.environ, self.config):
                if k in store:
                    value = store[k]
                    if boolean:
                        if value in ("yes", "no"):
                            value = value == "yes"
                        else:
                            continue
                    if k in compat_keys:
                        replacements = f"{self.prefix}_{key}"
                        if key in self.defaults:
                            replacements += f"or UBUNTUTOOLS_{key}"
                        Logger.warning(
                            "Using deprecated configuration variable %s. You should use %s.",
                            k,
                            replacements,
                        )
                    return value
        return default


def ubu_email(name=None, email=None, export=True):
    """Find the developer's Ubuntu e-mail address, and export it in
    DEBFULLNAME, DEBEMAIL if necessary (and export isn't False).

    e-mail Priority: arguments, UBUMAIL, DEBEMAIL, EMAIL, user@mailname
    name Priority: arguments, UBUMAIL, DEBFULLNAME, DEBEMAIL, NAME, /etc/passwd

    Name and email are only exported if provided as arguments or found in
    UBUMAIL. Otherwise, wrapped devscripts scripts can be expected to determine
    the values themselves.

    Return name, email.
    """
    name_email_re = re.compile(r"^\s*(.+?)\s*<(.+@.+)>\s*$")

    if email:
        match = name_email_re.match(email)
        if match and not name:
            name = match.group(1)
            email = match.group(2)

    if export and not name and not email and "UBUMAIL" not in os.environ:
        export = False

    for var, target in (
        ("UBUMAIL", "email"),
        ("DEBFULLNAME", "name"),
        ("DEBEMAIL", "email"),
        ("EMAIL", "email"),
        ("NAME", "name"),
    ):
        if name and email:
            break
        if var in os.environ:
            match = name_email_re.match(os.environ[var])
            if match:
                if not name:
                    name = match.group(1)
                if not email:
                    email = match.group(2)
            elif target == "name" and not name:
                name = os.environ[var].strip()
            elif target == "email" and not email:
                email = os.environ[var].strip()

    if not name:
        gecos_name = pwd.getpwuid(os.getuid()).pw_gecos.split(",")[0].strip()
        if gecos_name:
            name = gecos_name

    if not email:
        mailname = socket.getfqdn()
        if os.path.isfile("/etc/mailname"):
            mailname = open("/etc/mailname", "r", encoding="utf-8").read().strip()
        email = f"{pwd.getpwuid(os.getuid()).pw_name}@{mailname}"

    if export:
        os.environ["DEBFULLNAME"] = name
        os.environ["DEBEMAIL"] = email

    # decode env var or gecos raw string with the current locale's encoding
    encoding = locale.getlocale()[1]
    if not encoding:
        encoding = "utf-8"
    if name and isinstance(name, bytes):
        name = name.decode(encoding)
    return name, email



#
# main.py - main function for sponsor-patch script
#
# Copyright (C) 2010, Benjamin Drung <bdrung@ubuntu.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import logging
import os
import pwd
import shutil
import subprocess
import sys

from distro_info import UbuntuDistroInfo
from launchpadlib.launchpad import Launchpad

from ubuntutools.question import input_number
from ubuntutools.sponsor_patch.bugtask import BugTask, is_sync
from ubuntutools.sponsor_patch.patch import Patch
from ubuntutools.sponsor_patch.question import ask_for_manual_fixing
from ubuntutools.sponsor_patch.source_package import SourcePackage
from ubuntutools.update_maintainer import MaintainerUpdateException, update_maintainer

Logger = logging.getLogger(__name__)


def is_command_available(command, check_sbin=False):
    "Is command in $PATH?"
    path = os.environ.get("PATH", "/usr/bin:/bin").split(":")
    if check_sbin:
        path += [f"{directory[:-3]}sbin" for directory in path if directory.endswith("/bin")]
    return any(os.access(os.path.join(directory, command), os.X_OK) for directory in path)


def check_dependencies():
    "Do we have all the commands we need for full functionality?"
    missing = []
    for cmd in ("patch", "quilt", "dput", "lintian"):
        if not is_command_available(cmd):
            missing.append(cmd)
    if not any(
        is_command_available(cmd, check_sbin=True) for cmd in ("pbuilder", "sbuild", "cowbuilder")
    ):
        missing.append("pbuilder/cowbuilder/sbuild")

    if missing:
        Logger.warning(
            "sponsor-patch requires %s to be installed for full functionality", ", ".join(missing)
        )


def get_source_package_name(bug_task):
    package = None
    if bug_task.bug_target_name != "ubuntu":
        assert bug_task.bug_target_name.endswith("(Ubuntu)")
        package = bug_task.bug_target_name.split(" ")[0]
    return package


def get_user_shell():
    try:
        shell = os.environ["SHELL"]
    except KeyError:
        shell = pwd.getpwuid(os.getuid())[6]
    return shell


def edit_source():
    # Spawn shell to allow modifications
    cmd = [get_user_shell()]
    Logger.debug(" ".join(cmd))
    print(
        f"""An interactive shell was launched in
file://{os.getcwd()}
Edit your files. When you are done, exit the shell. If you wish to abort the
process, exit the shell such that it returns an exit code other than zero.
""",
        end=" ",
    )
    returncode = subprocess.call(cmd)
    if returncode != 0:
        Logger.error("Shell exited with exit value %i.", returncode)
        sys.exit(1)


def ask_for_patch_or_branch(bug, attached_patches, linked_branches):
    patch = None
    branch = None
    if len(attached_patches) == 0:
        msg = f"{len(linked_branches)} branches linked:"
    elif len(linked_branches) == 0:
        msg = f"{len(attached_patches)} patches attached:"
    else:
        branches = f"{len(linked_branches)} branch"
        if len(linked_branches) > 1:
            branches += "es"
        patches = f"{len(attached_patches)} patch"
        if len(attached_patches) > 1:
            patches += "es"
        msg = f"{branches} linked and {patches} attached:"
    Logger.info("https://launchpad.net/bugs/%i has %s", bug.id, msg)
    i = 0
    for linked_branch in linked_branches:
        i += 1
        print(f"{i}) {linked_branch.display_name}")
    for attached_patch in attached_patches:
        i += 1
        print(f"{i}) {attached_patch.title}")
    selected = input_number("Which branch or patch do you want to download", 1, i, i)
    if selected <= len(linked_branches):
        branch = linked_branches[selected - 1].bzr_identity
    else:
        patch = Patch(attached_patches[selected - len(linked_branches) - 1])
    return (patch, branch)


def get_patch_or_branch(bug):
    patch = None
    branch = None
    if not is_sync(bug):
        attached_patches = [a for a in bug.attachments if a.type == "Patch"]
        linked_branches = [b.branch for b in bug.linked_branches]
        if len(attached_patches) == 0 and len(linked_branches) == 0:
            if len(bug.attachments) == 0:
                Logger.error(
                    "No attachment and no linked branch found on "
                    "bug #%i. Add the tag sync to the bug if it is "
                    "a sync request.",
                    bug.id,
                )
            else:
                Logger.error(
                    "No attached patch and no linked branch found. "
                    "Go to https://launchpad.net/bugs/%i and mark an "
                    "attachment as patch.",
                    bug.id,
                )
            sys.exit(1)
        elif len(attached_patches) == 1 and len(linked_branches) == 0:
            patch = Patch(attached_patches[0])
        elif len(attached_patches) == 0 and len(linked_branches) == 1:
            branch = linked_branches[0].bzr_identity
        else:
            patch, branch = ask_for_patch_or_branch(bug, attached_patches, linked_branches)
    return (patch, branch)


def download_branch(branch):
    dir_name = os.path.basename(branch)
    if os.path.isdir(dir_name):
        shutil.rmtree(dir_name)
    cmd = ["bzr", "branch", branch]
    Logger.debug(" ".join(cmd))
    if subprocess.call(cmd) != 0:
        Logger.error("Failed to download branch %s.", branch)
        sys.exit(1)
    return dir_name


def merge_branch(branch):
    edit = False
    cmd = ["bzr", "merge", branch]
    Logger.debug(" ".join(cmd))
    if subprocess.call(cmd) != 0:
        Logger.error("Failed to merge branch %s.", branch)
        ask_for_manual_fixing()
        edit = True
    return edit


def extract_source(dsc_file, verbose=False):
    cmd = ["dpkg-source", "--skip-patches", "-x", dsc_file]
    if not verbose:
        cmd.insert(1, "-q")
    Logger.debug(" ".join(cmd))
    if subprocess.call(cmd) != 0:
        Logger.error("Extraction of %s failed.", os.path.basename(dsc_file))
        sys.exit(1)


def get_open_ubuntu_bug_task(launchpad, bug, branch=None):
    """Returns an open Ubuntu bug task for a given Launchpad bug.

    The bug task needs to be open (not complete) and target Ubuntu. The user
    will be ask to select one if multiple open Ubuntu bug task exits for the
    bug.
    """
    bug_tasks = [BugTask(x, launchpad) for x in bug.bug_tasks]
    ubuntu_tasks = [x for x in bug_tasks if x.is_ubuntu_task()]
    bug_id = bug.id
    if branch:
        branch = branch.split("/")
        # Non-production LP?
        if len(branch) > 5:
            branch = branch[3:]

    if len(ubuntu_tasks) == 0:
        Logger.error("No Ubuntu bug task found on bug #%i.", bug_id)
        sys.exit(1)
    elif len(ubuntu_tasks) == 1:
        task = ubuntu_tasks[0]
    elif branch and branch[1] == "ubuntu":
        tasks = [t for t in ubuntu_tasks if t.get_series() == branch[2] and t.package == branch[3]]
        if len(tasks) > 1:
            # A bug targeted to the development series?
            tasks = [t for t in tasks if t.series is not None]
        assert len(tasks) == 1
        task = tasks[0]
    else:
        task_list = [t.get_short_info() for t in ubuntu_tasks]
        Logger.debug(
            "%i Ubuntu tasks exist for bug #%i.\n%s",
            len(ubuntu_tasks),
            bug_id,
            "\n".join(task_list),
        )
        open_ubuntu_tasks = [x for x in ubuntu_tasks if not x.is_complete()]
        if len(open_ubuntu_tasks) == 1:
            task = open_ubuntu_tasks[0]
        else:
            Logger.info(
                "https://launchpad.net/bugs/%i has %i Ubuntu tasks:", bug_id, len(ubuntu_tasks)
            )
            for i, ubuntu_task in enumerate(ubuntu_tasks):
                print(f"{i + 1}) {ubuntu_task.get_package_and_series()}")
            selected = input_number(
                "To which Ubuntu task does the patch belong", 1, len(ubuntu_tasks)
            )
            task = ubuntu_tasks[selected - 1]
    Logger.debug("Selected Ubuntu task: %s", task.get_short_info())
    return task


def _create_and_change_into(workdir):
    """Create (if it does not exits) and change into given working directory."""

    if not os.path.isdir(workdir):
        try:
            os.makedirs(workdir)
        except os.error as error:
            Logger.error(
                "Failed to create the working directory %s [Errno %i]: %s.",
                workdir,
                error.errno,
                error.strerror,
            )
            sys.exit(1)
    if workdir != os.getcwd():
        Logger.debug("cd %s", workdir)
        os.chdir(workdir)


def _update_maintainer_field():
    """Update the Maintainer field in debian/control."""
    Logger.debug("update-maintainer")
    try:
        update_maintainer("debian", Logger.isEnabledFor(logging.DEBUG))
    except MaintainerUpdateException as e:
        Logger.error("update-maintainer failed: %s", str(e))
        sys.exit(1)


def _update_timestamp():
    """Run dch to update the timestamp of debian/changelog."""
    cmd = ["dch", "--maintmaint", "--release", ""]
    Logger.debug(" ".join(cmd))
    if subprocess.call(cmd) != 0:
        Logger.debug("Failed to update timestamp in debian/changelog.")


def _download_and_change_into(task, dsc_file, patch, branch):
    """Downloads the patch and branch and changes into the source directory."""

    if branch:
        branch_dir = download_branch(task.get_branch_link())

        # change directory
        Logger.debug("cd %s", branch_dir)
        os.chdir(branch_dir)
    else:
        if patch:
            patch.download()

        Logger.debug("Ubuntu package: %s", task.package)
        if task.is_merge():
            Logger.debug("The task is a merge request.")
        if task.is_sync():
            Logger.debug("The task is a sync request.")

        extract_source(dsc_file, Logger.isEnabledFor(logging.DEBUG))

        # change directory
        directory = f"{task.package}-{task.get_version().upstream_version}"
        Logger.debug("cd %s", directory)
        os.chdir(directory)


def sponsor_patch(bug_number, build, builder, edit, keyid, lpinstance, update, upload, workdir):
    workdir = os.path.realpath(os.path.expanduser(workdir))
    _create_and_change_into(workdir)

    launchpad = Launchpad.login_with("sponsor-patch", lpinstance)
    bug = launchpad.bugs[bug_number]

    patch, branch = get_patch_or_branch(bug)
    task = get_open_ubuntu_bug_task(launchpad, bug, branch)

    dsc_file = task.download_source()

    _download_and_change_into(task, dsc_file, patch, branch)

    source_package = SourcePackage(task.package, builder, workdir, branch)

    if is_sync(bug) and not edit:
        successful = source_package.reload_changelog()

        if successful:
            source_package.check_sync_request_version(bug_number, task)
            previous_version = task.get_previous_version()
            successful = source_package.check_version(previous_version)

        if successful:
            if build:
                dist = UbuntuDistroInfo().devel()
                successful = source_package.build(update, dist)
                update = False
            else:
                # We are going to run lintian, so we need a source package
                successful = source_package.build_source(None, upload, previous_version)

        if successful:
            series = task.get_debian_source_series()
            if source_package.sync(upload, series, bug_number, bug.owner.name):
                return
        edit = True

    if patch:
        edit |= patch.apply(task)
    elif branch:
        edit |= merge_branch(branch)

    while True:
        if edit:
            edit_source()
        # All following loop executions require manual editing.
        edit = True

        _update_maintainer_field()
        if not source_package.reload_changelog():
            continue

        if not source_package.check_version(task.get_version()):
            continue

        _update_timestamp()

        if not source_package.build_source(keyid, upload, task.get_previous_version()):
            continue

        source_package.generate_debdiff(dsc_file)

        # Make sure that the Launchpad bug will be closed
        if not source_package.is_fixed(bug):
            continue

        if not source_package.check_target(upload, launchpad):
            continue

        if build:
            successful_built = source_package.build(update)
            update = False
            if not successful_built:
                continue

        if not source_package.ask_and_upload(upload):
            continue

        # Leave while loop if everything worked
        break
